//
///  <ID>$Id: cmdopt.h,v 1.4 2003/04/28 07:11:54 t Exp $</ID>
//
///  <COLLECTION>MakLib Code Library</COLLECTION>
//
///  <name>cmdopt.h</name>
//
///  <overview>
///  <jpn>R}hCIvV</jpn>
///  <eng>Command-line option processing </eng>
///  </overview>
//
//    Copyright (c) 2003 Makino Takaki, t-libmak@snowelm.com
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#ifndef	__cmdopt_h
#define __cmdopt_h

#include <ctype.h>
#include <limits.h>

#include <string>
#include <vector>
#include <map>
#include <algorithm>
#if defined(LIBMAK_HAVE_HASH_MAP)
#include <hash_map>
#elif defined(LIBMAK_HAVE_EXT_HASH_MAP)
#include <ext/hash_map>
#endif
#ifdef LIBMAK_HAVE_SSTREAM
#include <sstream>
#endif

#include "except.h"
#include "ref.h"
#include "infinity.h"
#include "typestr.h"
#include "stringop.h"
#include "hash_string.h"

namespace mak {

class cmdopt_parser;	// forward declaration


#if defined(LIBMAK_HAVE_HASH_IN_GNU)
using __gnu_cxx::hash_map;
#elif defined(LIBMAK_HAVE_HASH_IN_STD)
using std::hash_map;
#endif


/// <notiondef>
///
/// <name> cmdopt_overview </name>
/// <title> R}hCIvV[` </title>
/// 
/// <desc>
/// 
/// <p> <ref>cmdopt.h</ref> ɂ́AR}hC̈Alo߂̃CuĂ܂B</p>
/// <ul>
/// <li> <ref>cmdopt_parser</ref> ^̃IuWFNg p pӂBꂪ͊ɂȂ܂B
/// <li> <ref>cmdopt</ref> NXA邢͂ق <ref>cmdopt_base</ref> ^pNX
///      A <ref>ref</ref> ɂQƂ p.add ŉĂ܂B
/// <li> p.parse(argc, argv) 邢 p.parse(std::istream&) ƂƁAR}hC߂܂B
/// <li> IvV̒l͊e cmdopt ^ get \bhŒׂ܂BIvVłȂ
///       p.argv() œ邱Ƃł܂B
/// </ul>
/// 
/// <samp>
/// #include &lt;iostream&gt;<br>
/// #include &lt;vector&gt;<br>
/// #include "cmdopt.h"<br>
/// <br>
/// int main(int argc, char **argv)<br>
/// {<br>
/// @@<ref>cmdopt_parser</ref> p;<br>
/// <br>
/// @@cmdopt&lt;int&gt; cnt('c', "count", 0, <br>
/// @@@@"Counts something.");<br>
/// @@cmdopt&lt;std::string&gt; set('s', "set", "", <br>
/// @@@@"Sets something.");<br>
/// @@cmdopt&lt;bool&gt; bit('b', "bit", false, <br>
/// @@@@"Flags something.");<br>
/// <br>
/// @@p.add( ref(cnt) );<br>
/// @@p.add( ref(set) );<br>
/// @@p.add( ref(bit) );<br>
/// @@p.add( New&lt;<ref>cmdopt_help</ref>&gt;('h', "help", p,<br>
/// @@@@"Show this help.") );<br>
/// @@p.add( New&lt;<ref>cmdopt_version</ref>&gt;('v', "version", "Command Line Option Tester 1.0",<br>
/// @@@@"Show the version of this software.") );<br>
/// <br>
/// @@try {<br>
/// @@@@p.parse(argc, argv);<br>
/// @@}<br>
/// @@catch(exception &e)<br>
/// @@{<br>
/// @@@@cout &lt;&lt; "Errors in options: " &lt;&lt; endl &lt;&lt; e.what();<br>
/// @@@@return 1;<br>
/// @@}<br>
/// <br>
/// @@cout &lt;&lt; "cnt value: " &lt;&lt; count.get() &lt;&lt; endl;<br>
/// @@cout &lt;&lt; "set value: " &lt;&lt; set.get() &lt;&lt; endl;<br>
/// @@cout &lt;&lt; "bit value: " &lt;&lt; bit.get() &lt;&lt; endl;<br>
/// <br>
/// @@vector<std::string> s = p.argv();<br>
/// @@cout &lt;&lt; "non-option arguments: " &lt;&lt; endl;<br>
/// @@for( vector<std::string>::iterator i = s.begin(); i != s.end(); i++ )<br>
/// @@{<br>
/// @@@@cout &lt;&lt; "'" &lt;&lt; *i &lt;&lt; "'" &lt;&lt; endl;<br>
/// @@}<br>
/// @@return 0;<br>
/// }<br>
/// </samp>
/// </desc>
/// <see> cmdopt, cmdopt_parser, cmdopt_help</see>
/// </notiondef>


/////////////////////////////////////////////////////////////////////////

/// <classdef>
/// <name> exception_cmdopt </name>

/// <overview> <jpn>R}hCONX</jpn><eng>Command-line option error exception</eng> </overview>
/// <desc>
/// <jpn>
/// <p>
/// R}hC IvV̏ɂăG[ƂɓONXłB
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// An exception class thrown by the error of command-line option processing.
/// </p>
/// </eng>
/// </desc>
/// <see> <ref>cmdopt_parser</ref> </see>

/// <body>

class exception_cmdopt : public exception_base
{
public:
	exception_cmdopt(std::string reason) : exception_base(reason) { }
		/// <jpn>R <var>reason</var> ɂċNꂽO܂B</jpn>
		/// <eng>Construct an exception with reason <var>reason</var>.</eng>
};

/// </body>
/// </classdef>


/////////////////////////////////////////////////////////////////////////

/// <classdef>
/// <name> cmdopt_base </name>

/// <overview> <jpn>R}hCIvV x[XNX</jpn><eng>Command-line option base class</eng> </overview>
/// <desc>
/// <jpn>
/// <p>
/// R}hC IvV\x[XƂȂ钊ۃNXłBׂẴR}hC IvV NX
/// ̃NXh܂B
/// ʏ́ilƂjIvVł΁A<ref>cmdopt</ref> gΑvłAȃIvV
/// Ƃ͂̃NXhĂBh̗ <ref>cmdopt_help</ref> QƂĂB
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// </p>
/// </eng>
/// </desc>
/// <see> <ref>cmdopt</ref>, <ref>cmdopt_help</ref>, <ref>cmdopt_version</ref> </see>

/// <body>

class cmdopt_base
{
private:
	unsigned char opt;
		/// One-character option
	std::string name;
		/// Long option 

protected:
	std::string _usage;
		/// Usage std::string

	virtual std::string usage_valstr() const { return ""; }

public:
	cmdopt_base(std::string iusage, unsigned char iopt, std::string iname) : opt(iopt), name(iname), _usage(iusage) { }
		/// 1IvVAIvVAp@w肵č\z܂B
	cmdopt_base(std::string iusage,                     std::string iname) : opt('\0'), name(iname), _usage(iusage) { }
		/// IvVƗp@w肵č\z܂B1IvV͎g܂B
	virtual ~cmdopt_base() { }
		/// gpĂ郁܂B<ref>cmdopt_parser</ref> ɓo^ĂԂɌĂ΂Ă͂܂B

	std::string longopt() const { return name; }
		/// IvV擾܂B
	unsigned char charopt() const{ return opt; }
		/// 1IvV擾܂B

	virtual bool parse(std::istream& is, char delim = '\0') = 0;
		/// ̃IvVR}hCɎw肳ꂽƂ cmdopt_parser Ăяo鏃z֐łB
		/// ʏ́Ǎ̒l <var>is</var> 璊oď܂B
		/// <var>delim</var> ͂̃IvV̋؂蕶ł ('\0' ̂Ƃɂ isspace() ɊY镶ׂċ؂蕶ɂȂ܂)B<br>
		/// ߂l <var>delim</var> 𒊏oĂȂƂ^łB
	virtual void usage(std::ostream &o) const;
		/// ̃IvV̎gp@ <var>o</var> ɏo͂܂B

	virtual void setParser(cmdopt_parser &) { }

	virtual std::string makeString() const { return ""; }
		/// ̃IvV̒l𕶎Ƃĕ\܂B
};
/// </body>
/// </classdef>


/////////////////////////////////////////////////////////////////////////

/// <classdef>
/// <name> cmdopt </name>

/// <overview> <jpn>R}hCIvV NX</jpn><eng>Command-line option class</eng> </overview>
/// <desc>
/// <jpn>
/// <p>
/// ʓIȒlƂR}hC IvṼNXłB
/// ev[g̈ɂ́AIvVƂĎľ^w肵܂B
/// int, std::string, double, bool ɑ΂Ă͓ʃCX^Xo[WpӂĂ܂B
/// ȊÔ̂w肷ꍇŁAW̒oZqł͕s\ȏꍇɂ̓[UʃCX^Xo[W`ĂB
/// ܂Ǎ^w肷ꍇŁAʂ̓ꍇ́Acmdopt NXpƕ֗łB
/// ̑́A蕡GȃIvVɂĂ <ref>cmdopt_base</ref> hĂB
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// </p>
/// </eng>
/// </desc>
/// <see> <ref>cmdopt_base</ref>, <ref>cmdopt_parser</ref> </see>

/// <body>

template <class T>
class cmdopt : public cmdopt_base
{
public:
	typedef T val_t;
		/// lƂĂƂ^łB

private:
	val_t val;
	val_t defval;

public:
	cmdopt(std::string iusage, unsigned char iopt, std::string iname, const val_t &ival = val_t()) : cmdopt_base(iusage, iopt, iname), val(ival), defval(ival) { }
		/// 1IvVAIvVAftHglAp@w肵č\z܂B
	cmdopt(std::string iusage, std::string iname, const val_t &ival = val_t()) : cmdopt_base(iusage, iname), val(ival), defval(ival) { }
		/// IvVAftHglAp@w肵č\z܂B

	virtual const val_t & get() const { return val; }
		/// l擾܂B
	virtual void set(const val_t &t) { val = t; }
		/// lݒ肵܂B
	virtual const val_t & getDefval() const { return defval; }
		/// ftHgl擾܂B

	operator const val_t &() const { return get(); }

	virtual bool parse(std::istream& is, char delim);
		/// IvV^ꂽƂǍ̒l <var>is</var> 璊oď܂B

	virtual std::string makeString() const { if( val == defval ) return ""; else return "--" + longopt() + "=" + makestr(val); }

protected:
	virtual std::string usage_valstr() const { return "=" + typestring<T>(); };
		/// IvV̌ɏƂ "=..." Ԃ܂Busage() Ă΂܂B
};
/// </body>
/// </classdef>

/// <!-- cmdopt templatized member function -->

template <class T>
bool cmdopt<T>::parse(std::istream& is, char delim)
{
	char c;
	val_t v;
	bool delim_kept = true;
	
	if( is.eof() )
		throw exception_cmdopt("No value specified for option --" + longopt());
	else 
	{
		is.get(c);
		if( isspace(c) || c == delim )
		{
			delim_kept = false;
			if( c != delim )
				while( isspace(c) )
					is.get(c);
			else
				is.get(c); //// todo: ̂ւ̂II
			if( isdigit(c) || c == '-' )
			{
				is.putback(c);
				is >> v;
				delim_kept = true;
			}
			else
				throw exception_cmdopt("Value required for option --" + longopt());
		}
		else if( c == '=' )
			is >> v;
			else
		{
			is.putback(c);
			if( isdigit(c) )
				is >> v;
			else
				throw exception_cmdopt("Value required for option --" + longopt());
		}
	}
	if( is.good() )
		set(v);
	else
		throw exception_cmdopt("Bad value for option --" + longopt());

	return delim_kept;
}

/// <!-- user specialization classes of cmdopt -->

template <> class cmdopt<int> : public cmdopt_base
{
	typedef int val_t;
	
	val_t val;
	val_t min;
	val_t max;
	val_t defval;

public:
	cmdopt(std::string iusage, unsigned char iopt, std::string iname, val_t ival = 0, val_t imin = INT_MIN, val_t imax = INT_MAX) : cmdopt_base(iusage, iopt, iname), val(ival), min(imin), max(imax), defval(ival) { }
		/// 1IvVAIvVAftHglAp@w肵č\z܂B
	cmdopt(std::string iusage, std::string iname, val_t ival = 0, val_t imin = INT_MIN, val_t imax = INT_MAX) : cmdopt_base(iusage, iname), val(ival), min(imin), max(imax), defval(ival) { }
		/// IvVAftHglAp@w肵č\z܂B

	virtual const val_t & get() const { return val; }
	virtual void set(const val_t &t) { val = t; }
	operator const val_t &() const { return get(); }
	virtual const val_t & getDefval() const { return defval; }

	virtual bool parse(std::istream& is, char delim);

	virtual std::string makeString() const { if( val == defval ) return ""; else return "--" + longopt() + "=" + makestr(val); }

protected:
	virtual std::string usage_valstr() const { return "=integer"; };
};

template <> class cmdopt<std::string> : public cmdopt_base
{
	typedef std::string val_t;
	val_t val;
	val_t defval;

public:
	cmdopt(std::string iusage, unsigned char iopt, std::string iname, val_t ival = "") : cmdopt_base(iusage, iopt, iname), val(ival), defval(ival) { }
	cmdopt(std::string iusage, std::string iname, val_t ival = "") : cmdopt_base(iusage, iname), val(ival), defval(ival) { }

	virtual const val_t & get() const { return val; }
	virtual void set(const val_t &t) { val = t; }
	operator const val_t &() const { return get(); }
	virtual const val_t & getDefval() const { return defval; }

	virtual bool parse(std::istream& is, char delim);

	virtual std::string makeString() const { if( val == defval ) return ""; else return "--" + longopt() + "=" + quotestr(val); }

protected:
	virtual std::string usage_valstr() const { return "=std::string"; };
};

template <> class cmdopt<bool> : public cmdopt_base
{
	typedef bool val_t;
	val_t val;
	val_t defval;

public:
	cmdopt(std::string iusage, unsigned char iopt, std::string iname, val_t ival = false) : cmdopt_base(iusage, iopt, iname), val(ival), defval(ival) { }
	cmdopt(std::string iusage, std::string iname, val_t ival = false) : cmdopt_base(iusage, iname), val(ival) { }

	virtual const val_t & get() const { return val; }
	virtual void set(const val_t &t) { val = t; }
	operator const val_t &() const { return get(); }
	virtual const val_t & getDefval() const { return defval; }

	virtual bool parse(std::istream& is, char delim);

	virtual std::string makeString() const { if( val == defval ) return ""; else return "--" + longopt() + "=" + makestr(val); }

protected:
	virtual std::string usage_valstr() const { return "[+|-]"; };
};

template <> class cmdopt<double> : public cmdopt_base
{
	typedef double val_t;
	val_t val;
	val_t min;
	val_t max;
	val_t defval;

public:
	cmdopt(std::string iusage, unsigned char iopt, std::string iname, val_t ival = 0.0, val_t imin = -Infinity, val_t imax = Infinity) : cmdopt_base(iusage, iopt, iname), val(ival), min(imin), max(imax), defval(ival) { }
		/// 1IvVAIvVAftHglAp@w肵č\z܂B
	cmdopt(std::string iusage, std::string iname, val_t ival = 0.0, val_t imin = -Infinity, val_t imax = Infinity) : cmdopt_base(iusage, iname), val(ival), min(imin), max(imax), defval(ival) { }
		/// IvVAftHglAp@w肵č\z܂B

	virtual const val_t & get() const { return val; }
	virtual void set(const val_t &t) { val = t; }
	operator const val_t &() const { return get(); }
	virtual const val_t & getDefval() const { return defval; }

	virtual bool parse(std::istream& is, char delim);

	virtual std::string makeString() const { if( val == defval ) return ""; else return "--" + longopt() + "=" + makestr(val); }

protected:
	virtual std::string usage_valstr() const { return "=float"; };
};

/////////////////////////////////////////////////////////////////////////

/// <classdef>
/// <name> cmdopt_help </name>

/// <overview> <jpn>R}hC wvIvVNX</jpn><eng>Command-line help option class</eng> </overview>
/// <desc>
/// <jpn>
/// <p>
/// wvƂăR}hC IvV̈ꗗ\IvVłB\z̎ɁA
/// IvVꗗn <ref>cmdopt_parser</ref> w肵܂B
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// </p>
/// </eng>
/// </desc>
/// <see> <ref>cmdopt_base</ref>, <ref>cmdopt_parser</ref> </see>

/// <body>

class cmdopt_help : public cmdopt_base
{
	const cmdopt_parser *parser;
public:
	cmdopt_help(std::string iusage, unsigned char iopt, std::string iname) : cmdopt_base(iusage, iopt, iname) { parser = NULL; }
		/// 1IvVAIvVAcmdopt_parserAp@w肵č\z܂B
	cmdopt_help(std::string iusage, std::string iname) : cmdopt_base(iusage, iname) { parser = NULL; }
		/// IvVAcmdopt_parserAp@w肵č\z܂B

	virtual void setParser(cmdopt_parser &p) { parser = &p; }
	bool parse(std::istream& is, char delim);
		/// ̃IvVw肳ꂽƂ cmdopt_parser Ăяo܂BR}hꗗ cout ɕ\ exit(1) ܂B

protected:
	virtual std::string usage_valstr() const { return ""; };
		/// IvVɗ^鐔l\ƂāA󕶎Ԃ܂B
};
/// </body></classdef>


/////////////////////////////////////////////////////////////////////////

/// <classdef>
/// <name> cmdopt_version </name>

/// <overview> <jpn>R}hC o[W\IvVNX</jpn><eng>Command-line version option class</eng> </overview>
/// <desc>
/// <jpn>
/// <p>
/// \tgEFAA쌠\Ao[W\IvVłB\z̎ɁA
/// \w肵܂B
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// </p>
/// </eng>
/// </desc>
/// <see> <ref>cmdopt_base</ref>, <ref>cmdopt_parser</ref> </see>

/// <body>

class cmdopt_version : public cmdopt_base
{
	std::string version;

public:
	cmdopt_version(std::string iusage, unsigned char iopt, std::string iname, std::string ver) : cmdopt_base(iusage, iopt, iname), version(ver) { }
		/// 1IvVAIvVAo[WAp@w肵č\z܂B
	cmdopt_version(std::string iusage, std::string iname, std::string ver) : cmdopt_base(iusage, iname), version(ver) { }
		/// IvVAo[WAp@w肵č\z܂B

	bool parse(std::istream& is, char delim);
		/// ̃IvVw肳ꂽƂ cmdopt_parser Ăяo܂Bo[W cout ɕ\ exit(1) ܂B

protected:
	virtual std::string usage_valstr() const { return ""; };
		/// IvVɗ^鐔l\ƂāA󕶎Ԃ܂B
};
/// </body></classdef>


/// <classdef>
/// <name> cmdopt_dummy </name>

/// <overview> <jpn>R}hC wvێ\NX</jpn><eng>Command-line help std::string class</eng> </overview>
/// <desc>
/// <jpn>
/// <p>
/// wv\̒Ŏw肵\܂Bʏ̃IvVł͂Ȃ̂ŁAIvVȂǂ͎wł܂B
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// A dummy option class to insert the specified std::string in the usage screen.
/// </p>
/// </eng>
/// </desc>
/// <see> <ref>cmdopt_base</ref>, <ref>cmdopt_parser</ref> </see>

/// <body>

class cmdopt_dummy : public cmdopt_base
{

public:
	cmdopt_dummy(std::string iusage) : cmdopt_base(iusage, '\0', "") { }
		/// p@w肵č\z܂B

	bool parse(std::istream&, char) { throw exception_cmdopt("Internal error: Dummy option specified"); }
		/// ̃IvVw肳ꂽƂ cmdopt_parser Ăяo܂Bo[W cout ɕ\ exit(1) ܂B

	void SetUsage(std::string s) { _usage = s; }

protected:
	virtual void usage(std::ostream &o) const { o << _usage << std::endl; }
		/// IvVɗ^鐔l\ƂāA󕶎Ԃ܂B
};
/// </body></classdef>



template <class T, class VEC_ENTRY = T>
class cmdopt_vec : public cmdopt<T>
{
	typedef typename cmdopt<T>::val_t val_t;
	cmdopt_parser *p;
	std::vector<VEC_ENTRY> &vec; 
public:
	cmdopt_vec(std::string iusage, unsigned char iopt, std::string iname, std::vector<VEC_ENTRY> &ivec, const val_t &ival = val_t()) 
		: cmdopt<T>(iusage, iopt, iname, ival), p(NULL), vec(ivec) { }
		/// 1IvVAIvVAftHglAp@w肵č\z܂B
	cmdopt_vec(std::string iusage, std::string iname, std::vector<VEC_ENTRY> &ivec, const val_t &ival = val_t()) 
		: cmdopt<T>(iusage, iname, ival), p(NULL), vec(ivec) { }
		/// IvVAftHglAp@w肵č\z܂B
	template <class T1>
	cmdopt_vec(std::string iusage, unsigned char iopt, std::string iname, std::vector<VEC_ENTRY> &ivec, const val_t &ival, const T1 &v1) 
		: cmdopt<T>(iusage, iopt, iname, ival, v1), p(NULL), vec(ivec) { }
	template <class T1>
	cmdopt_vec(std::string iusage, std::string iname, std::vector<VEC_ENTRY> &ivec, const val_t &ival, const T1 &v1) 
		: cmdopt<T>(iusage, iname, ival, v1), p(NULL), vec(ivec) { }
	template <class T1, class T2>
	cmdopt_vec(std::string iusage, unsigned char iopt, std::string iname, std::vector<VEC_ENTRY> &ivec, const val_t &ival, const T1 &v1, const T2 &v2) 
		: cmdopt<T>(iusage, iopt, iname, ival, v1, v2), p(NULL), vec(ivec) { }
	template <class T1, class T2>
	cmdopt_vec(std::string iusage, std::string iname, std::vector<VEC_ENTRY> &ivec, const val_t &ival, const T1 &v1, const T2 &v2) 
		: cmdopt<T>(iusage, iname, ival, v1, v2), p(NULL), vec(ivec) { }

	virtual bool parse(std::istream& is, char delim) { bool ret = cmdopt<T>::parse(is, delim); vec.push_back( VEC_ENTRY(*this) ); return ret;}
		/// IvV^ꂽƂǍ̒l <var>is</var> 璊oď܂B

	virtual void setParser(cmdopt_parser &parser) { p = &parser; }
	cmdopt_parser &getParser() { if( p == NULL ) throw exception_cmdopt("Parser not set"); return *p; }

	virtual std::string makeString() const { std::string ret = ""; std::string delim = ""; for( typename std::vector<VEC_ENTRY>::iterator i = vec.begin(); i != vec.end(); i++ ) { ret = ret + delim + "--" + longopt() + "=" + makestr(*i); delim = " "; } return ret; }
};

/////////////////////////////////////////////////////////////////////////

/// <classdef>
/// <name> cmdopt_parser </name>

/// <overview> <jpn>R}hCIvVNX</jpn><eng>Command-line option processor</eng> </overview>
/// <desc>
/// <jpn>
/// <p>
/// R}hĈ߂̃NXłB
/// X̃IvV\NX (<ref>cmdopt</ref> ^Âق <ref>cmdopt_base</ref> ^hNX)pӂāA
/// add() őgݍ񂾌Aparse() ĂяoăIvV܂B
/// ʂ́AeIvVNX get() œǂݏoAargv() Ŏo肵܂B
/// ȂIvVꍇ͗O <ref>exception_cmdopt</ref> o܂B
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// </p>
/// </eng>
/// </desc>
/// <see> <ref>cmdopt_base</ref>, <ref>cmdopt</ref>, <ref>exception_cmdopt</ref> </see>

/// <body>

class cmdopt_parser 
{
	typedef hash_map<std::string, ref<cmdopt_base> > strmap_t;
	typedef hash_map<unsigned char,  ref<cmdopt_base> > charmap_t;
	typedef std::vector< std::pair< ref<cmdopt_base>, bool> >  vec_t;
	
	strmap_t strmap;
	charmap_t charmap;
	vec_t vec;

	std::vector<std::string> _argv;

public:
	cmdopt_parser() { }
			/// IvVȂNX܂B
	~cmdopt_parser() { }
			/// gĂ (<ref>ref</ref> NXʂ) ܂B

	void add( ref<cmdopt_base> t, bool show = true ) 
		{ 
		  std::string optname = t->longopt();
		  strmap_t::iterator i = strmap.find(optname); 
		  if( i != strmap.end() ) { erase(i->second); } 
		  if( optname != "" )
			  strmap[optname] = t;
		  if( t->charopt() != '\0' )
			  charmap[t->charopt()] = t; 
		  vec.push_back(std::make_pair(t,show)); 
		  t->setParser(*this); 
		}
			/// R}hC IvVǉ܂Bshow  false w肷ƁAwvŕ\܂B
	void erase( const ref<cmdopt_base> t ) { 
		if( t->longopt() != "" )
		{
			strmap_t::iterator i = strmap.find(t->longopt());
			if( i != strmap.end() && i->second == t )
				strmap.erase(i);
		}
		if( t->charopt() != '\0' )
		{
			charmap_t::iterator i = charmap.find(t->charopt());
			if( i != charmap.end() && i->second == t )
				charmap.erase(i);
		}
		vec_t::iterator i;
		for( i=vec.begin(); i!=vec.end(); i++) 
			if( i->first == t ) { vec.erase(i); break;}
	}
			/// R}hC IvV܂B

	int parse(std::istream& is, char delim = '\0') throw( exception_cmdopt );
			/// <var>is</var>  <var>delim</var> ŋ؂ꂽi󔒕̏ꍇ <var>delim</var>='\0'j
			/// IvV𒊏oď܂B
	int parse(int argc, char **argv) throw( exception_cmdopt );
			/// main() ɗ^`̃IvV܂B
	int parse(std::string s, char delim = '\0') throw( exception_cmdopt ) { std::istringstream iss(s); return parse(iss, delim); }
			/// ̒̃IvV܂B
	void usage(std::ostream& os) const;
			/// ׂẴIvVɂĂ̎gp@ os ɑ}܂B

	std::vector<std::string> & argv() { return _argv; }
	const std::vector<std::string> & argv() const { return _argv; }
			///	parse() ĂяoAIvVłȂR}hC̈ꗗ擾邽߂̊֐łB

	std::string makeString() const;
};

/// </body>
/// </classdef>

} // namespace mak

#endif	// __cmdopt_h
