//  $Id: initial.cpp,v 1.2 2003/04/28 05:49:15 t Exp $
//
///  MakLib Code Library
//
//  initial.cpp
//
///  <overview>
///  <jpn>/I</jpn>
///  <eng>Initialization and Termination</eng>
///  </overview>
//
//    Copyright (c) 2003 Makino Takaki, t-libmak@snowelm.com
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

static const char rcsid[] = "$Id: initial.cpp,v 1.2 2003/04/28 05:49:15 t Exp $";

#include "initial.h"
#include <iostream>

namespace mak {

InitTermBase::InitTermBase(InitTermBase **table, int ip)
{
	priority = ip;
	while(*table != NULL && (*table)->priority <= ip) 
	{
		table = &((*table)->next);
	}
	
	next = *table;
	*table = this;
}

void InitTermBase::PerformAll(InitTermBase *table)
{
	while( table != NULL )
	{
		table->Work();
		table = table->next;
	}
}

Initializer::Initializer(int ip) 
	: InitTermBase( &initable, ip )
{
	
}

Terminator::Terminator(int ip) 
	: InitTermBase( &trmtable, -ip )
{
	
}

void Initializer::PerformAll()
{
	InitTermBase::PerformAll(initable); 
}

void Terminator::PerformAll()
{
	InitTermBase::PerformAll(trmtable); 
}

InitTermBase *Initializer::initable = NULL;
InitTermBase *Terminator::trmtable = NULL;


InitFunction::InitFunction(void (*f)(), int ip)
	: Initializer(ip) { func = f; } 

void InitFunction::Work() 
{ 
	(*func)();
}

TermFunction::TermFunction(void (*f)(), int ip)
	: Terminator(ip) { func = f; } 

void TermFunction::Work() 
{ 
	(*func)();
}


} // namespace mak
