//    $Id: libmaktest.cpp,v 1.3 2003/04/28 05:49:15 t Exp $
//
//    libmaktest.cpp: Test program for libmak
//
//    Copyright (c) 2003 Makino Takaki, t-libmak@snowelm.com
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//
#include <iostream>

#include "config_maklib.h"
#include "test.h"
#include "cmdopt.h"
#include "ref.h"

using namespace std;
using namespace mak;

TestSuite cmdoptsuite("cmdopt");
TestSuite testtestsuite("test");

class TestTest : public TestBundle
{
public:
	TestTest() : TestBundle(&testtestsuite) { }
	virtual int testCount() { return 6; }
	virtual void test(int i) { 
		switch(i)
		{
			case 1:
				success("Always success"); 
				break;
			case 2:
				testEqual(3+4, 7);
				break;
			case 3:
				testEqualDelta(1.0/3.0, 0.33333333, 1e-5);
				break;
			case 4:
				testThrow(throw 3, int);
				break;
			case 5:
				testNoThrow(while(0) { });
				break;
			case 6:
				failureExpect(); failure("Expected to be failed");
				break;
		}
	}
} testtest;

class TestFailTest : public TestBundle
{
public:
	TestFailTest() : TestBundle(&testtestsuite) { }

	virtual void success(std::string desc = "Success") throw() { TestBase::failure(desc); }
	virtual void failure(std::string desc = "Failure") throw() { TestBase::success(desc); }
};

class TestFailMainTest : public TestFailTest
{
public:
	TestFailMainTest() : TestFailTest() { }

	virtual int testCount() { return 6; }
	virtual void test(int i) { 
		switch(i)
		{
			case 1:
				failure("Test of failure action (expected to be failed)");
				break;
			case 2:
				testEqual(3+4, 8);
				break;
			case 3:
				testEqualDelta(1.0/3.0, 0.333, 1e-5);
				break;
			case 4:
				testThrow(throw 3.0, int);
				break;
			case 5:
				testThrow(while(0) { }, int);
				break;
			case 6:
				testNoThrow(throw 3);
				break;
		}
	}
} testfailmaintest;

class TestFailSetupTest : public TestFailTest
{
public:
	TestFailSetupTest() : TestFailTest() { }

	virtual void setUp() {
		throw "setUp throws something";
	}

	virtual int testCount() { return 1; }
	virtual void test(int) { 
		TestBase::failure("Failed to catch exception");
	}
} testfailsetuptest;

class TestFailThrowTest : public TestFailTest
{
public:
	TestFailThrowTest() : TestFailTest() { }

	virtual int testCount() { return 1; }
	virtual void test(int) { 
		throw "test throws something";
	}
} testfailthrowtest;


class TestFailTearTest : public TestFailTest
{
public:
	TestFailTearTest() : TestFailTest() { }

	virtual int testCount() { return 1; }

	virtual void tearDown() {
		throw "tearDown throws something";
	}

	virtual void test(int) { 
	}
} testfailteardowntest;


class CmdOptTest : public TestBundle
{
protected:
	cmdopt_parser *p;
	cmdopt<int> *flg_a;
	cmdopt<std::string> *flg_b;
	cmdopt<bool> *flg_c;
	cmdopt<double> *flg_d;

	CmdOptTest() : TestBundle(&cmdoptsuite) { }
	
	virtual void setUp() { p = new cmdopt_parser; 
	    flg_a = new cmdopt<int>("Counts something.", 'a', "count", 1234, -9999, 9999);
	    flg_b = new cmdopt<string>("Sets something.", 'b', "set", "Hogehoge");
	    flg_c = new cmdopt<bool>("Flags something.", 'c', "bit", false);
	    flg_d = new cmdopt<double>("Doubles something.", 'd', "double", 1234.0, -9999.0, 9999.0);

	    p->add( Ref(*flg_a) );
	    p->add( Ref(*flg_b) );
	    p->add( Ref(*flg_c) );
	    p->add( Ref(*flg_d) );
	    p->add( New<cmdopt_help>("Show this help.", 'h', "help") );
	    p->add( New<cmdopt_version>("Show the version of this software.", 'v', "version", "Command Line Option Tester 1.0") );
	}
	virtual void tearDown() { delete p; delete flg_a; delete flg_b; delete flg_c; delete flg_d; }
};

class CmdOptIntegerTest : public CmdOptTest
{
public:
	virtual int testCount() { return 19; }
	virtual void test(int i) { 
		switch(i) {
		case 1:
			testEqual(flg_a->get(),1234);
			break;
		case 2:
			testNoThrow(p->parse("-a123"));
			testEqual(flg_a->get(),123);
			break;
		case 3:
			testNoThrow(p->parse("-a\n129\n",'\n'));
			testEqual(flg_a->get(),129);
			break;
		case 4:
			testNoThrow(p->parse("-a-123"));
			testEqual(flg_a->get(),-123);
			break;
		case 5:
			testNoThrow(p->parse("-a\n-129\n",'\n'));
			testEqual(flg_a->get(),-129);
			break;
		case 6:
			testNoThrow(p->parse("-a=124"));
			testEqual(flg_a->get(),124);
			break;
		case 7:
			testNoThrow(p->parse("-a=-124"));
			testEqual(flg_a->get(),-124);
			break;
		case 8:
			testNoThrow(p->parse("--count\n121\n",'\n'));
			testEqual(flg_a->get(),121);
			break;
		case 9:
			testNoThrow(p->parse("--count\n-121\n",'\n'));
			testEqual(flg_a->get(),-121);
			break;
		case 10:
			testNoThrow(p->parse("--count=127"));
			testEqual(flg_a->get(),127);
			break;
		case 11:
			testNoThrow(p->parse("--count=-127"));
			testEqual(flg_a->get(),-127);
			break;
		case 12:
		    p->add( mak::New< cmdopt<int> >("Ones something.", '1', "one", 1234, -9999, 9999));
			testNoThrow(p->parse("-a\n-1\n",'\n'));
			testEqual(flg_a->get(),-1);
			break;
		case 13:
			testThrow(p->parse("-a10000"),exception_cmdopt);
			break;
		case 14:
			testThrow(p->parse("-a-10000"),exception_cmdopt);
			break;
		case 15:
			testThrow(p->parse("-a"),exception_cmdopt);
			break;
		case 16:
			testThrow(p->parse("-a "),exception_cmdopt);
			break;
		case 17:
			testThrow(p->parse("-a\n",'\n'),exception_cmdopt);
			break;
		case 18:
			testThrow(p->parse("-a \n",'\n'),exception_cmdopt);
			break;
		case 19:
			testThrow(p->parse("-a\n-c\n",'\n'),exception_cmdopt);
			break;
		}
	}
} cmdoptintegertest;

class CmdOptDoubleTest : public CmdOptTest
{
public:
	virtual int testCount() { return 14; }
	virtual void test(int i) { 
		switch(i) {
		case 1:
			testEqual(flg_d->get(),1234.0);
		break;
		case 2:
			testNoThrow(p->parse("-d123"));
			testEqual(flg_d->get(),123.0);
		break;
		case 3:
			testNoThrow(p->parse("-d\n1.0e+2\n",'\n'));
			testEqual(flg_d->get(),100.0);
		break;
		case 4:
			testNoThrow(p->parse("-d-123"));
			testEqual(flg_d->get(),-123.0);
		break;
		case 5:
			testNoThrow(p->parse("-d\n-129\n",'\n'));
			testEqual(flg_d->get(),-129.0);
		break;
		case 6:
			testNoThrow(p->parse("-d=1.345"));
			testEqual(flg_d->get(),1.345);
		break;
		case 7:
			testNoThrow(p->parse("-d=-1.333e-1"));
			testEqual(flg_d->get(),-0.1333);
		break;
		case 8:
			testNoThrow(p->parse("--double\n121\n",'\n'));
			testEqual(flg_d->get(),121.0);
		break;
		case 9:
			testNoThrow(p->parse("--double\n-0.3e+1\n",'\n'));
			testEqual(flg_d->get(),-3.0);
		break;
		case 10:
			testNoThrow(p->parse("--double\n.3e-1\n",'\n'));
			testEqual(flg_d->get(),0.03);
		break;
		case 11:
			testNoThrow(p->parse("--double=127"));
			testEqual(flg_d->get(),127.0);
		break;
		case 12:
			testNoThrow(p->parse("--double=-12e0"));
			testEqual(flg_d->get(),-12.0);
		break;
		case 13:
			testThrow(p->parse("-d10000"),exception_cmdopt);
		break;
		case 14:
			testThrow(p->parse("-d-10000"),exception_cmdopt);
		break;
		}
	}
} cmdoptdoubletest;

class CmdOptStringTest : public CmdOptTest
{
public:
	virtual int testCount() { return 23; }
	virtual void test(int i) { 
		switch(i) {
		case 1:
			testEqual(flg_b->get(),"Hogehoge");
		break;
		case 2:
			testNoThrow(p->parse("-b123"));
			testEqual(flg_b->get(),"123");
		break;
		case 3:
			testNoThrow(p->parse("-b\n\n-c\n",'\n'));
			testEqual(flg_b->get(),"");
		break;
		case 4:
			testNoThrow(p->parse("-bhogera\n-c-\n",'\n'));
			testEqual(flg_b->get(),"hogera");
		break;
		case 5:
			testNoThrow(p->parse("-b\nhogehoge\n-c-\n",'\n'));
			testEqual(flg_b->get(),"hogehoge");
		break;
		case 6:
			testNoThrow(p->parse("-bHOGE"));
			testEqual(flg_b->get(),"HOGE");
		break;
		case 7:
			testNoThrow(p->parse("-b=hoge"));
			testEqual(flg_b->get(),"hoge");
		break;
		case 8:
			testNoThrow(p->parse("-b=hoga hoge"));
			testEqual(flg_b->get(),"hoga");
		break;
		case 9:
			testNoThrow(p->parse("--set 123"));
			testEqual(flg_b->get(),"123");
		break;
		case 10:
			testThrow(p->parse("--set\n-c\n",'\n'), exception_cmdopt);
		break;
		case 11:
			testNoThrow(p->parse("--set\nhogehoge\n-c\n",'\n'));
			testEqual(flg_b->get(),"hogehoge");
		break;
		case 12:
			testNoThrow(p->parse("--set\nHOGE",'\n'));
			testEqual(flg_b->get(),"HOGE");
		break;
		case 13:
			testNoThrow(p->parse("--set\nHOGA\n",'\n'));
			testEqual(flg_b->get(),"HOGA");
		break;
		case 14:
			testNoThrow(p->parse("--set=hoge"));
			testEqual(flg_b->get(),"hoge");
		break;
		case 15:
			testNoThrow(p->parse("--set=\n",'\n'));
			testEqual(flg_b->get(),"");
		break;
		case 16:
			testNoThrow(p->parse("--set='hoge hoge'"));
			testEqual(flg_b->get(),"hoge hoge");
		break;
		case 17:
			testNoThrow(p->parse("-b="));
			testEqual(flg_b->get(),"");
		break;
		case 18:
			testNoThrow(p->parse("-b= \n",'\n'));
			testEqual(flg_b->get()," ");
		break;
		case 19:
			testNoThrow(p->parse("-b= "));
			testEqual(flg_b->get(),"");
		break;
		case 20:
			testNoThrow(p->parse("-b='hoge hoge'"));
			testEqual(flg_b->get(),"hoge hoge");
		break;
		case 21:
			testThrow(p->parse("--set 123",'\n'),exception_cmdopt);
		break;
		case 22:
			testThrow(p->parse("--set",'\n'),exception_cmdopt);
		break;
		case 23:
			testThrow(p->parse("--set\n",'\n'),exception_cmdopt);
		break;
		}
	}
} cmdoptstringtest;

class CmdOptBoolTest : public CmdOptTest
{
public:
	virtual int testCount() { return 10; }
	virtual void test(int i) { 
		switch(i) {
		case 1:
			testEqual(flg_c->get(),false);
		break;
		case 2:
			testNoThrow(p->parse("-c"));
			testEqual(flg_c->get(),true);
		break;
		case 3:
			testNoThrow(p->parse("-c-"));
			testEqual(flg_c->get(),false);
		break;
		case 4:
			testNoThrow(p->parse("-c+"));
			testEqual(flg_c->get(),true);
		break;
		case 5:
			testNoThrow(p->parse("-c0"));
			testEqual(flg_c->get(),false);
		break;
		case 6:
			testNoThrow(p->parse("-c1"));
			testEqual(flg_c->get(),true);
		break;
		case 7:
			testNoThrow(p->parse("-c=false"));
			testEqual(flg_c->get(),false);
		break;
		case 8:
			testNoThrow(p->parse("-c=true"));
			testEqual(flg_c->get(),true);
		break;
		case 9:
			testNoThrow(p->parse("-c=no"));
			testEqual(flg_c->get(),false);
		break;
		case 10:
			testNoThrow(p->parse("-c=yes"));
			testEqual(flg_c->get(),true);
		break;
		}
	}
} cmdoptbooltest;

class CmdOptFailTest : public CmdOptTest
{
public:
	virtual int testCount() { return 5; }
	virtual void test(int i) { 
		switch(i) {
		case 1:
			testThrow(p->parse("-u"),exception_cmdopt);
		break;
		case 2:
			testThrow(p->parse("--unknown"),exception_cmdopt);
		break;
		case 3:
			testThrow(p->parse("-b -u"),exception_cmdopt);
		break;
		case 4:
			testThrow(p->parse("-ub"),exception_cmdopt);
		break;
		case 5:
			testNoThrow(p->parse("-bu"));
		break;
		}
	}
} cmdoptfailtest;


int main(int argc, char **argv)
{
	std::cout << "Test program for " << PACKAGE_NAME << " " << PACKAGE_VERSION << std::endl;
	TestSuite::setVerboseAll(1);
	return TestSuite::testAll();
}

