#! /usr/local/bin/perl
# $Id: makedoc.perl,v 1.2 2003/04/28 05:49:15 t Exp $
#
#    makedoc.perl: automatic document generator
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


$docdir = "doc/";

@order = ("typedef", "subclass", "enum", 
	"const", "member",
	"ctor", "dtor", 
	"func", "operator", 
	"staticmember", "staticfunc",
	"class"
);

$linkstr{"Index"} = "<jpn></jpn><eng>Index</eng>";
$link{"Index"} = "index.html";

for( $pass = 1; $pass <= 2; $pass++ )
{
	$membercount = 1;
	
	print STDERR "Pass $pass...\n";

	foreach $f (@ARGV)
	{
		$tmp = $f;
		$tmp =~ s/\./_/;
		$link{$f} = "$tmp.html";
		$inbuf = "";
		$title = "";
		$namespace = "";

		open(IN, "<$f");
		while( <IN> )
		{
			chop;

			if( /namespace (\w[\w0-9]*)\s*($|{)/ ) ## }
			{
				$namespace = $1;
			}

			if( m#^\s*///\s(.*)$# )
			{
				$l = $1;
				if( $l =~ /<(class|func)def>/ ) {
					$l = $l . "<namespace>$namespace</namespace>"; }
				$inbuf = $inbuf . $l . "\n";
				$title = $l if( $title eq "" );
			}
			elsif( m#^(.*[^/])///\s(.*)$# )
			{
				($a,$b) = ($1, $2);
				$a = &encode(&escape($a));
				$inbuf = $inbuf . "<!-- " . $a . " -->" . $b . "\n";
			}
			elsif( m#\S# )	# non-space char
			{
				$a = &encode(&escape($_));
				$inbuf = $inbuf . "<!-- " . $a . " -->\n";
			}
			else
			{
	#			$inbuf = $inbuf . "\n";
			}
		}
		close(IN);

		@outn = @outc = @outf = @outv = @outk = ();
		while( $inbuf =~ m#<notiondef>(.*?)</notiondef>#sig )
		{
			push( @outn, &notiondef($1, $f, $title) );
		}
		while( $inbuf =~ m#<classdef>(.*?)</classdef>#sig )
		{
			push( @outc, &classdef($1, $f, $title)) ;
		}
		while( $inbuf =~ m#<funcdef>(.*?)</funcdef>#sig )
		{
			push( @outf,  &funcdef($1, $f, $title) );
		}
		while( $inbuf =~ m#<vardef>(.*?)</vardef>#sig )
		{
			push( @outv,  &vardef($1, $f, $title) );
		}
		while( $inbuf =~ m#<constdef>(.*?)</constdef>#sig )
		{
			push( @outk,  &constdef($1, $f, $title) );
		}

		if( $inbuf =~ m#<overview>(.*?)</overview>#si )
		{
			$h_overview = " - $1";
		}
		else
		{
			$h_overview = "";
		}

		if( $pass == 2 )
		{
			$out = "<H1><jpn>wb_t@C</jpn><eng>Header file</eng> $f $h_overview</H1>";
			if( @outn != 0 )
			{
				$out = $out . "<H2>Tv</H2>\n<ul>" . join("\n", @outn) . "</ul>";
			}
			if( @outc != 0 )
			{
				$out = $out . "<H2>NX</H2>\n<ul>" . join("\n", @outc) . "</ul>";
			}
			if( @outf != 0 )
			{
				$out = $out . "<H2>֐</H2>\n<ul>" . join("\n", @outf) . "</ul>";
			}
			if( @outk != 0 )
			{
				$out = $out . "<H2>萔</H2>\n<ul>" . join("\n", @outk) . "</ul>";
			}
			if( @outv != 0 )
			{
				$out = $out . "<H2>ϐ</H2>\n<ul>" . join("\n", @outv) . "</ul>";
			}

			$out = $out . "<H2><jpn>Q</jpn><eng>See also</eng></H2>\n<ref>Index</ref>\n";

			&sendfile( $link{$f}, "Header file $f: $title", $out);

			push( @allouth, "<li><ref>$f</ref>$h_overview" );
			push( @alloutn, @outn );
			push( @alloutc, @outc );
			push( @alloutf, @outf );
			push( @alloutv, @outv );
			push( @alloutk, @outk );
		}
	}
}

print STDERR "Finish...\n";

$out = "<H1>S̍</H1>\n";
$out = $out . "<H2>wb_t@C</H2>\n<ul>" . join("\n", sort @allouth) . "</ul>\n";
$out = $out . "<H2>Tv</H2>\n<ul>" . join("\n", sort @alloutn) . "</ul>\n";
$out = $out . "<H2>NX</H2>\n<ul>" . join("\n", sort @alloutc) . "</ul>\n";
$out = $out . "<H2>֐</H2>\n<ul>" . join("\n", sort @alloutf) . "</ul>\n";
$out = $out . "<H2>萔</H2>\n<ul>" . join("\n", sort @alloutk) . "</ul>\n";
$out = $out . "<H2>ϐ</H2>\n<ul>" . join("\n", sort @alloutv) . "</ul>\n";

&sendfile( "index.html", "Index", $out);


sub notiondef
{
	local($str, $file) = @_;
	local($name, $desc, $see, $decl, $namespace, $classhead, $body, $out, $lin, $delim, $mode );
	if( $str =~ s#<name>(.*?)</name>##si )
	{
		$name = $1;
		$name =~ s/\s//g;
	}
	else { die "no <name> in notiondef in file $file" };

	if( $str =~ s#<title>(.*?)</title>##si ) { $title = $1; }	else { die "no <title> in notiondef $name in file $file" };
	if( $str =~ s#<desc>(.*?)</desc>##si ) { $desc = $1; } else { die "no <desc> in notiondef $name in file $file" };

	if( $str =~ s#<see>(.*?)</see>##si ) { $see = &makeref($1); }		else { $see = "" };

	if( $pass == 1 )
	{
		$link{$name} = "$name.html";
		$linkstr{$name} = $title;
	}
	else
	{
		$out = "<H1>$title</H1>\n";

		$out = $out . "$desc\n";

		$out = $out . "<H2><jpn>Q</jpn><eng>See also</eng></H2>\n$see\n" if $see ne "";

# ŏI

		$out =~ s#\n\n\n+#\n\n#g;

#		print $out;
		&sendfile($link{$name}, "Function $name: $title", $out);
	}

	return "<li><ref>$name</ref>";
}

sub classdef
{
	local($str, $file, $title) = @_;
	local($name, $desc, $see, $decl, $namespace, $classhead, $body, $out, $lin, $delim, $mode );

	if( $str =~ s#<name>(.*?)</name>##si )
	{
		$name = $1;
		$name =~ s/\s//g;
	}
	else { die "no <name> in classdef" };

	if( $str =~ s#<desc>(.*?)</desc>##si )
	{
		$desc = $1;
	}
	else { die "no <desc> in classdef $name" };

	if( $str =~ s#<remark>(.*?)</remark>##si ) { $remark = $1; }	else { $remark = "" };
	if( $str =~ s#<overview>(.*?)</overview>##si ) { $overview = " - $1"; }	else { $overview = "" };
	if( $str =~ s#<see>(.*?)</see>##si ) { $see = &makeref($1); }		else { $see = "" };
	if( $str =~ s#<decl>(.*?)</decl>##si ) { $decl = $1; }	else { $decl = "" };
	if( $str =~ s#<namespace>(.*?)</namespace>##si ) { $namespace = $1; }	else { $namespace = "" };


	if( $str =~ s#<body>(.*?)</body>##si )
	{
		$body = $1;
		$body =~ s# -->\n<!-- #\n#g;

# find class header
		if( !( $body =~ s#^[\s\n]*<!-- \s*(([^{;&]|&[^;]+;)*)[{;]##s ) )
		{
			die "no class header found at the top in classdef $name";
		}
		else
		{
			$classhead = $1;
			$classhead =~ s#^[\s\n]+##;		# Remove head spaces
			$classhead =~ s#[\s\n]+$##;		# Remove tail spaces
			$classhead =~ s#\n#<br>\n#g;		# Insert <br>
			if( $classhead =~ s/([^:])\s*:([^:].*)/$1/ )
			{ 
				if( $pass == 1 )
				{
# make reference to parent classes 
					$derivation = $2;
					foreach $i ( split(/,/, $derivation) )
					{
						$i =~ s/^\s+//g;
						$i =~ s/\s+$//g;
						$i =~ s/\s+/ /g;
						($t,$x,$x,$n) = ($i =~ /((virtual )?(public|private|protected)) (.*)/);
						if( $t eq "public" ) { $t = "";	}
							else { $t = " ($t)"; }

						push(@{$parent_info{$name}}, $n);
						push(@{$parent_class{$name}}, $n . $t);
						push(@{$child_class{$n}}, $name . $t);
					}
				}
			}
			$classhead = $classhead . ";";
			if( $namespace ne "" )
			{
				$classhead = "namespace $namespace {\n<div style=\"margin-left: 2em\">$classhead</div>\n}";
			}
			$decl = "#include \"<ref>" . $file . "</ref>\"<br><br>\n" . $classhead if( $decl eq "" );
		}

		undef %func;
		$mode = "private";
		$mode = "public" if( $decl =~ /struct/ );

		if( $decl !~ /\btypedef\b/ )
		{
			while(1)
			{
				if( $body =~ s/^[\s\n]*(private|public|protected)\s*:// )
				{	$mode = $1; next; }
				last if ! ($body =~ s#^[\s\n]*(([^;{}&]|&[a-z]+;)*?)([;{}])##s);
				($lin, $delim) = ($1,$3);
				last if( $delim eq "}" );
#				print "$lin <<<$delim>>>\n";
				if( $delim eq "{" )
				{
					while( $body =~ /([{}])/ )
					{
	#					print "--------\n $body";
						if( $1 eq "}" )
						{
							$body =~ s/^[^}]*\}//s;
	#						print "--------\n";
							last;
						}
						if( ! ($body =~ s/\{[^\{\}]*\}//s) )
						{
							die "why?"
						}
					}
				}
				if( $lin ne "" )
				{
					$fname = "";
					$descr = "";
					if( $lin =~ s/[\s\n]*-->(.*?)<!--[\s\n]*$//s )
					{
						$descr = $descr . $1;
					}
					while( $body =~ s/^[\s\n]*-->(.*?)<!--[\s\n]*//s )
					{
						$descr = $descr . $1;
					}

					if( $lin =~ /(^|\W)friend\W/ )
					{
						$section = "friend func";
						$section = "friend class" if $lin =~ /(^|\W)(class|struct)\W/;
						$section = "friend operator" if $lin =~ /\Woperator\W/;

						if( $section eq "friend func" )
						{
							($dummy,$fname) = ($lin =~ /^(|[^(]+[^(\w])(\w+)\(/);
						}
						elsif( $section eq "friend operator" )
						{
							($fname) = ($lin =~ /(operator([^(]|\(.*\))+)\(/);
							$fname =~ s/operator\s*/operator /;
						}
					}
					elsif( $lin =~ /(^|^[^\(]*)~$name\(/ )
					{
						$section = "$mode dtor";
					}
					elsif( $lin =~ /(^|^[^\(]*[^\(\w])$name\(/ )
					{
						$section = "$mode ctor";
					}
					elsif( $lin =~ /typedef/ )
					{
						$section = "$mode typedef";
						($fname) = ($lin =~ /[^\w0-9](\w[\w0-9]+)[^\w0-9]*$/);
					}
					elsif( $lin =~ /(^|^[^\(]*[^\(\w])enum\W/ )
					{
						$section = "$mode enum";
						($fname) = ($lin =~ /enum\s+(\w+)/);
					}
					elsif( $lin =~ /(^|^[^\(]*[^\(\w])(class|struct)\W/ )
					{
						$section = "$mode subclass";
						($dummy,$fname) = ($lin =~ /(class|struct)\s+(\w+)/);
					}
					elsif( $lin =~ /(^|\W)operator\W/ )
					{
						$section = "$mode operator";
						($fname) = ($lin =~ /(operator([^(]|\(.*\))+)\(/);
						$fname =~ s/operator\s*/operator /;
					}
					elsif( $lin =~ /^[^\(]*[A-Za-z0-9_]\(/ )
					{
						$section = "$mode func";
						$section = "$mode staticfunc" if( $lin =~ /(^|^.*\W)static\W/ );
						($dummy,$fname) = ($lin =~ /^(|[^(]+[^(\w])(\w+)\(/);
					}
					elsif( $lin =~ /(^|\W)static\W/ )
					{
						$section = "$mode staticmember";
						if( $lin =~ /(^|\W)(const\s+static|static\s+const)\W/ )
						{
							$section = "$mode const";
						}
						($fname) = ($lin =~ /[^\w0-9](\w[\w0-9]+)[^\w0-9]*$/);
					}
					else
					{
						$section = "$mode member";
						($fname) = ($lin =~ /[^\w0-9](\w[\w0-9]+)[^\w0-9]*$/);
					}

					if( $fname ne "" )
					{
						$fx = $fname;
						$fx =~ s/([\[\]\*\+])/\\$1/g;
						$fx =~ s/([A-Za-z0-9_])\s*([^A-Za-z0-9_\s])/$1\\s*$2/g;
						$fx =~ s/\s+/$1\\s+$2/g;
						$fx =~ s/^([A-Za-z_])/\\b$1/;
						$fx =~ s/([A-Za-z0-9_])$/$1\\b/;
						# print "$fx\n";
						$lin =~ s#$fx#<strong>$fname</strong>#;

						$fname =~ s/\bconst //;
					}
					$lin =~ s/\)[\s\n]*:.*/\)/sg;
					$lin =~ s/\s+$//s;
					$fdecl{$membercount} = $lin;
					$fsect{$membercount} = $section;
					$fdescr{$membercount} = $descr;
					$fname{$membercount} = $fname;
					$fclass{$membercount} = $name;
					push(@{$func{$section}}, $membercount);
					if( $pass == 1 )
					{
						push( @{$classmember{$name}{$fname}}, $membercount );
					}
					$membercount++;
				}
			}

			if( $pass == 2 && defined $parent_class{$name} )
			{
				&parent_search( $name, "root", $parent_class{$name} );
			}
		}
	}


	if( $pass == 1 )
	{
		$link{$name} = "$name.html";
		$link{$name} =~ s/([A-Z])/$1\^/g;
	}
	if( $pass == 2 )
	{
#		print "$name\n";
		$out = "<H1><eng>Class </eng>$name<jpn> NX</jpn>$overview</H1>\n";

		$out = $out . "<H2><jpn>錾</jpn><eng>Declaration</eng></H2>\n"
			. "<div><code>\n" . &makeref($decl, $name) . "</code></div>\n";

		if( defined $parent_class{$name} )
		{
			$out = $out . "<H2><jpn>eNX</jpn><eng>Parent classes</eng></H2>\n<p>";
			$connect = "";
			foreach $i (@{$parent_class{$name}})
			{
				$out = $out . $connect . &makeref($i, $name);
				$connect = ", ";
			}
			$out = $out . "</p>\n";
		}
		if( defined $child_class{$name} )
		{
			$out = $out . "<H2><jpn>qNX</jpn><eng>Child classes</eng></H2>\n<p>";
			$connect = "";
			foreach $i (@{$child_class{$name}})
			{
				$out = $out . $connect . &makeref($i, $name);
				$connect = ", ";
			}
			$out = $out . "</p>\n";
		}

		$out = $out . "<H2><jpn></jpn><eng>Description</eng></H2>\n$desc\n";

		$member_header = 0;

		foreach $i (@order)
		{
			foreach $j ("public", "protected", "friend")
			{
				$secname = "$j $i";
				if( defined $func{$secname} )
				{
					if( $member_header == 0 )
					{
						$member_header = 1;
						$out = $out . "<H2><jpn>o</jpn><eng>Members</eng></H2>\n";
					}

					$out = $out . "<H3>" . &makesecname($secname) . "</H3>\n";
					$close = "";
					$prevfname = "!";
#					print "-- " . $secname . " @{$func{$secname}}\n";
				
					foreach $x (sort {$fname{$a} cmp $fname{$b}} @{$func{$secname}} )
					{
#						print "$x\n";
						$derivestr = "";
						$derivestr = "  (<eng>derived from </eng>$fclass{$x}<jpn>p</jpn>)" if( $fclass{$x} ne $name );
						if( $prevfname ne $fname{$x} )
						{
							$prevfname = $fname{$x};
							if( $prevfname ne "" ) { $out = $out . "$close<H4>$prevfname $derivestr</H4><DL>"; }
							else		   { $out = $out . "$close<DL>"; }
							$close = "</DL>\n";
						}
						$out = $out ."<DT>" . &makeref($fdecl{$x}, $name) . ";\n<DD>" . $fdescr{$x} . "\n";
					}
					$out = $out . "$close\n";
				}
			}
		}

		$out = $out . "<H2><jpn></jpn><eng>Remarks</eng></H2>\n$remark\n" if $remark ne "";

		$out = $out . "<H2><jpn>Q</jpn><eng>See also</eng></H2>\n$see\n" if $see ne "";

# ŏI

		$out =~ s#\n\n\n+#\n\n#g;

		&sendfile($link{$name}, "Class $name: $title", $out);
	}

	return "<li><ref>$name</ref>$overview";
}

sub funcdef
{
	local($str, $file, $title) = @_;
	local($name, $desc, $see, $decl, $namespace);

	if( $str =~ s#<name>(.*?)</name>##si )
	{
		$name = $1;
		$name =~ s/\s//g;
	}
	else { die "no <name> in funcdef" };

	if( $str =~ s#<desc>(.*?)</desc>##si )
	{
		$desc = $1;
	}
	else { die "no <desc> in funcdef $name" };

	if( $str =~ s#<remark>(.*?)</remark>##si ) { $remark = $1; }	else { $remark = "" };
	if( $str =~ s#<overview>(.*?)</overview>##si ) { $overview = " - $1"; }	else { $overview = "" };
	if( $str =~ s#<see>(.*?)</see>##si ) { $see = &makeref($1,$name); }		else { $see = "" };
	if( $str =~ s#<decl>(.*?)</decl>##si ) { $decl = $1; }	else { $decl = "" };
	if( $str =~ s#<args>(.*?)</args>##si ) { $args = $1; }	else { $args = "" };
	if( $str =~ s#<retval>(.*?)</retval>##si ) { $retval = $1; }	else { $retval = "" };
	if( $str =~ s#<namespace>(.*?)</namespace>##si ) { $namespace = $1; }	else { $namespace = "" };


	if( $str =~ s#<body>(.*?)</body>##si )
	{
		$body = $1;
		$body =~ s# -->\n<!-- #\n#g;
		if( $decl eq "" )
		{
			($decl) = ($body =~ /<!--\s+(.*[^\s])\s+-->/);
			$decl =~ s/\binline\b//;	# cut off inline
			$decl =~ s/\s*[;{].*//;		# cut off function body
			$decl = "#include \"<ref>" . $file . "</ref>\"<br><br>\n$decl";
#			print $decl;
		}
	}

	if( $pass == 1 )
	{
		$link{"$name"} = "$name.html";
		$link{$name} =~ s/([A-Z])/$1\^/g;
		$linkstr{"$name"} = "$name()";
		$link{"$name()"} = "$name.html";
		$link{"$name()"} =~ s/([A-Z])/$1\^/g;
		$linkstr{"$name()"} = "$name()";
	}
	if( $pass == 2 )
	{
		$out = "<H1><eng>Function </eng>$name<jpn> ֐</jpn>$overview</H1>\n";

		$out = $out . "<H2><jpn>錾</jpn><eng>Declaration</eng></H2>\n"
			. "<div><code>\n" . &makeref($decl, $name) . ";</code></div>\n";

		$out = $out . "<H2><jpn></jpn><eng>Description</eng></H2>\n$desc\n";

		$out = $out . "<H2><jpn></jpn><eng>Arguments</eng></H2>\n$args\n" if $args ne "";

		$out = $out . "<H2><jpn>߂l</jpn><eng>Return value</eng></H2>\n$retval\n" if $retval ne "";

		$out = $out . "<H2><jpn></jpn><eng>Remarks</eng></H2>\n$remark\n" if $remark ne "";

		$out = $out . "<H2><jpn>Q</jpn><eng>See also</eng></H2>\n$see\n" if $see ne "";

# ŏI

		$out =~ s#\n\n\n+#\n\n#g;

#		print $out;
		&sendfile($link{$name}, "Function $name: $title", $out);
	}

	return "<li><ref>$name()</ref>$overview";
}

sub vardef
{
	return &convardef("var", @_);
}

sub constdef
{
	return &convardef("const", @_);
}

sub convardef
{
	local($convar, $str, $file, $title) = @_;
	local($name, $desc, $see, $decl, $namespace);

	if( $str =~ s#<name>(.*?)</name>##si )
	{
		$name = $1;
		$name =~ s/\s//g;
	}
	else { die "no <name> in ${convar}def" };

	if( $str =~ s#<desc>(.*?)</desc>##si )
	{
		$desc = $1;
	}
	else { $desc = "" };

	if( $str =~ s#<remark>(.*?)</remark>##si ) { $remark = $1; }	else { $remark = "" };
	if( $str =~ s#<overview>(.*?)</overview>##si ) { $overview = " - $1"; }	else { $overview = "" };
	if( $str =~ s#<see>(.*?)</see>##si ) { $see = &makeref($1,$name); }		else { $see = "" };
	if( $str =~ s#<decl>(.*?)</decl>##si ) { $decl = $1; }	else { $decl = "" };
	if( $str =~ s#<namespace>(.*?)</namespace>##si ) { $namespace = $1; }	else { $namespace = "" };

	if( $str =~ s#<body>(.*?)</body>##si )
	{
		$body = $1;
		$body =~ s# -->\n<!-- #\n#g;
		if( $decl eq "" )
		{
			($decl) = ($body =~ /<!--\s+(.*[^\s])\s+-->/);
			$decl =~ s/\s*[=\(;].*//;		# cut off variable body
			$decl = "#include \"<ref>" . $file . "</ref>\"<br><br>\n$decl";
#			print $decl;
		}
	}

	if( $pass == 1 )
	{
		$link{$name} = "$name.html";
		$link{$name} =~ s/([A-Z])/$1\^/g;
	}
	if( $pass == 2 )
	{
		if( $convar eq "var" )
		{
			$out = "<H1><eng>Variable </eng>$name<jpn> ϐ</jpn>$overview</H1>\n";
		}
		else
		{
			$out = "<H1><eng>Constant </eng>$name<jpn> 萔</jpn>$overview</H1>\n";
		}

		$out = $out . "<H2><jpn>錾</jpn><eng>Declaration</eng></H2>\n"
			. "<div><code>\n" . &makeref($decl, $name) . ";</code></div>\n";

		$out = $out . "<H2><jpn></jpn><eng>Description</eng></H2>\n$desc\n";

		$out = $out . "<H2><jpn></jpn><eng>Arguments</eng></H2>\n$args\n" if $args ne "";

		$out = $out . "<H2><jpn>߂l</jpn><eng>Return value</eng></H2>\n$retval\n" if $retval ne "";

		$out = $out . "<H2><jpn></jpn><eng>Remarks</eng></H2>\n$remark\n" if $remark ne "";

		$out = $out . "<H2><jpn>Q</jpn><eng>See also</eng></H2>\n$see\n" if $see ne "";

# ŏI

		$out =~ s#\n\n\n+#\n\n#g;

#		print $out;
		&sendfile($link{$name}, "$convar $name: $title", $out);
	}

	return "<li><ref>$name</ref>$overview";
}


sub escape
{
	local($s) = @_;
	$s =~ s#/\*([^\*]|\*[^/])+\*/##gs;
	$s =~ s/"([^"\\]|\\.)+"/"..."/gs;
	$s =~ s/'([^"\\]|\\.)+'/'.'/gs;
	return $s;
}

sub encode
{
	local($s) = @_;
	$s =~ s/&/&amp;/g;
	$s =~ s/>/&gt;/g;
	$s =~ s/</&lt;/g;
	return $s;
}

sub decode
{
	local($s) = @_;
	$s =~ s/&gt;/>/g;
	$s =~ s/&lt;/</g;
	$s =~ s/&amp;/&/g;
	return $s;
}

sub makeref
{
	local($s, $x) = @_;
	local($t) = &decode($s);
	local($i);
	local(@l) = split(/[^\w0-9\.]+/, $t);
	foreach $i (sort {length($b) <=> length($a)} @l)
	{
		if( defined $link{$i} && $i ne $x) 
		{
			$s =~ s#(^|[^\w<>/]|[^<]/|((([^<][Rr]|[^Rr])[Ee]|^[Ee])[Ff]|[^Ff])>)$i([^\w]|$)#$1<ref>$i</ref>$5#g;
			$s =~ s#(^|[^\w<>/]|[^<]/|((([^<][Rr]|[^Rr])[Ee]|^[Ee])[Ff]|[^Ff])>)$i([^\w]|$)#$1<ref>$i</ref>$5#g;
		}
	}
	return $s;
}

sub makelink
{
	local($s) = @_;
	return '<a href="' . $link{$s} . '">' . $linkstr{$s} . '</a>' if defined $linkstr{$s};
	return '<a href="' . $link{$s} . '">' . $s . '</a>'           if defined $link{$s};
	warn "Undefined reference $s";
	return $s;
}

sub sendfile
{
	local($name, $title, $out) = @_;

	$out =~ s#<eng>(.*?)</eng>##sig;
	$out =~ s#<jpn>(.*?)</jpn>#$1#sig;
	$out =~ s#<ref>(.*?)</ref>#&makelink($1)#esig;

	open (OUT, ">$docdir" . "$name");

	print OUT "<HTML>\n";
	print OUT "<!-- This document is automatically generated by makedoc.perl. DO NOT EDIT -->\n";
	print OUT "<!-- For copyright information, see the original source code -->\n";
	print OUT "<HEAD><TITLE> $title </TITLE></HEAD><BODY>\n";
	print OUT $out;
	print OUT "</BODY></HTML>\n";
	close(OUT);
}

sub parent_search
{
	local( $name, $imode, $parents ) = @_;

	foreach $parent( @{$parents} )
	{
		($class, $mode) = ($parent =~ /^(.*) \(.*\)$/);
		if( $mode eq "" )
		{
			$class = $parent;
			$mode = "public";
		}

		$mode =~ s/virtual\s+//g;

		next if( $imode ne "root" && $mode eq "private" );

		$mode = "protected" if $mode eq "public" && $imode eq "protected";
		$mode = "private" if $imode eq "private";

#		print "-- $name <- '$parent'\n";
		next if( ! defined $classmember{$parent} );
		foreach $fname ( keys %{$classmember{$parent}} )
		{
			if( ! defined $classmember{$name}{$fname} )
			{
#				print "--- $name <- $parent : $fname\n";
				foreach $membercount ( @{$classmember{$parent}{$fname}} )
				{
					$fsect = $fsect{$membercount};
					next if( $fsect =~ /private/ );
					if( $mode ne "public" && $fsect=~ /public/ )
					{
						$fsect =~ s/public/$mode/;
					}
					elsif( $mode eq "private" && $fsect =~ /protected/ )
					{
						$fsect =~ s/protected/private/;
					}
					push(@{$func{$fsect}}, $membercount);                                                        
#					print " *** $name, $fsect, $membercount, @{$func{$fsect}}\n";
				}
			}
		}
		if( defined $parent_class{$parent} )
		{
			&parent_search( $name, $mode, $parent_class{$parent} );
		}
	}
}


sub makesecname
{
	local($s) = @_;

	$s =~ s#friend func#<eng>Friend function</eng><jpn>th֐</jpn>#;

	$s =~ s#public ##;
	$s =~ s#private #<eng>Private </eng><jpn>vCx[g </jpn>#;
	$s =~ s#protected #<eng>Protected </eng><jpn>ی</jpn>#;
	$s =~ s#friend #<eng>Friend </eng><jpn>th </jpn>#;

	$s =~ s#typedef#<eng>Typedefs</eng><jpn>typedef</jpn>#;
	$s =~ s#subclass#<eng>inner class</eng><jpn>NX</jpn>#;
	$s =~ s#enum#<eng>enum</eng><jpn>񋓌^</jpn>#;
	$s =~ s#const#<eng>constant</eng><jpn>萔</jpn>#;

	$s =~ s#member#<eng>member variable</eng><jpn>oϐ</jpn>#;
	$s =~ s#ctor#<eng>constructor</eng><jpn>RXgN^</jpn>#;
	$s =~ s#dtor#<eng>destructor</eng><jpn>fXgN^</jpn>#;
	$s =~ s#func#<eng>member function</eng><jpn>o֐</jpn>#;
	$s =~ s#operator#<eng>operator</eng><jpn>Zq</jpn>#;
	$s =~ s#staticmember#<eng>static member variable</eng><jpn>ÓIoϐ</jpn>#;
	$s =~ s#staticfunc#<eng>static member function</eng><jpn>ÓIo֐</jpn>#;
	$s =~ s#class#<eng>class</eng><jpn>NX</jpn>#;

	return $s;
}
