//
///  <ID>$Id: maxmin.h,v 1.4 2003/04/28 07:11:54 t Exp $</ID>
//
///  <COLLECTION>MakLib Code Library</COLLECTION>
//
///  <name>maxmin.h</name>
//
///  <overview>
///  <jpn>Max  Min ̒`</jpn>
///  <eng>Definition of Max and Min</eng>
///  </overview>
//
//    Copyright (c) 2003 Makino Takaki, t-libmak@snowelm.com
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#ifndef __maxmin_h
#define __maxmin_h

#include "config_maklib.h"

namespace mak {

template <class CMP, class VAL> inline
VAL ifless( const CMP &x, const CMP &y, const VAL &a, const VAL &b ) { return x < y ? a : b; }

typedef unsigned int uint;

#define CONST_FUNCTION
// __attribute__((__constant__))

/////////////////////////////////////////////////////////////////////////

/// <funcdef>
/// <name> Max </name>

/// <overview><jpn>2̒l̂傫Ԃ</jpn><eng>Maximum of the two values</eng></overview>
/// <desc>
/// <jpn>
/// <p>
/// ӂ̒l̂A傫Ԃ܂B
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// Return the larger of the two values.
/// </p>
/// </eng>
/// </desc>
/// <remark>
///   <jpn><p> Pentium Pro ȍ~ CPU (Celeron, Pentium II ܂) ̊g߂g
///         őEŏ̏ł܂Bconfigure --enable-ppro-inst w肷邱Ƃ
///         ̖߂LɂȂ܂B </p></jpn>
///   <eng><p> Pentium Pro and later (including Celeron, Pentium II) have better instructions
///         that handle maximum and minimum.  You can use the instructions by 
///         configure --enable-ppro-inst. </p></eng>
/// </remark>
/// <see>Min</see>
/// <body>

template <class T> inline
T Max( const T &a, const T &b ) CONST_FUNCTION { return ifless(b,a,a,b); }

template <class T> inline
T Max( const T &a, const T &b, const T &c ) CONST_FUNCTION { return Max( Max(a,b), c); }

/// </body></funcdef>

/////////////////////////////////////////////////////////////////////////

/// <funcdef>
/// <name> Min </name>

/// <overview><jpn>2̒l̂Ԃ</jpn><eng>Minimum of the two values</eng></overview>
/// <desc>
/// <jpn>
/// <p>
/// ӂ̒l̂A傫Ԃ܂B
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// Return the smaller of the two values.
/// </p>
/// </eng>
/// </desc>
/// <remark>
///   <jpn> Pentium Pro ȍ~ CPU (Celeron, Pentium II ܂) ̊g߂g
///         őEŏ̏ł܂Bconfigure --enable-ppro-inst w肷邱Ƃ
///         ̖߂LɂȂ܂B </jpn>
///   <eng> Pentium Pro and later (including Celeron, Pentium II) have better instructions
///         that handle maximum and minimum.  You can use the instructions by 
///         configure --enable-ppro-inst. </eng>
/// </remark>
/// <see>Max</see>

/// <body>

template <class T> inline
T Min( const T &a, const T &b ) CONST_FUNCTION { return ifless(a,b,a,b); }

template <class T> inline
T Min( const T &a, const T &b, const T &c ) CONST_FUNCTION { return Min( Min(a,b), c); }

/// </body></funcdef>

#if defined(LIBMAK_USE_COND_MOVE)

template <>
inline int ifless<int, int>( const int &x, const int &y, const int &a, const int &b ) CONST_FUNCTION { 
	int result;
	asm(
	"cmpl	%2, %1\n"
	"cmovnl	%4, %0\n"
	: "=r,r" (result) : "rm,r" (x), "ri,m" (y), "0,0" (a), "rm,rm" (b) );
	return result;
}

template <>
inline uint ifless<int, uint>( const int &x, const int &y, const uint &a, const uint &b ) CONST_FUNCTION { 
	uint result;
	asm(
	"cmpl	%2, %1\n"
	"cmovnl	%4, %0\n"
	: "=r,r" (result) : "rm,r" (x), "ri,m" (y), "0,0" (a), "rm,rm" (b) );
	return result;
}

template <>
inline uint ifless<int, double>( const int &x, const int &y, const double &a, const double &b ) CONST_FUNCTION { 
	double result;
	asm(
	"cmpl	%2, %1\n"
	"fcmovnb	%4, %0\n"
	: "=r,r" (result) : "rm,r" ((x + 0x80000000)), "ri,m" ((y + 0x80000000)), "0,0" (a), "r,r" (b) );
	return result;
}

template <>
inline int ifless<uint, int>( const uint &x, const uint &y, const int &a, const int &b ) CONST_FUNCTION { 
	int result;
	asm(
	"cmpl	%2, %1\n"
	"cmovae	%4, %0\n"
	: "=r,r" (result) : "rm,i" (x), "ri,rm" (y), "0,0" (a), "rm,rm" (b) );
	return result;
}

template <>
inline uint ifless<uint, uint>( const uint &x, const uint &y, const uint &a, const uint &b ) CONST_FUNCTION { 
	uint result;
	asm(
	"cmpl	%2, %1\n"
	"cmovae	%4, %0\n"
	: "=r,r" (result) : "rm,i" (x), "ri,rm" (y), "0,0" (a), "rm,rm" (b) );
	return result;
}

template <>
inline double ifless<uint, double>( const uint &x, const uint &y, const double &a, const double &b ) CONST_FUNCTION { 
	double result;
	asm(
	"cmpl	%2, %1\n"
	"fcmovnb	%4, %0\n"
	: "=r" (result) : "rm" (x), "rc" (y), "r" (a), "r" (b) );
	return result;
}

template <>
inline int ifless<double, int>( const double &x, const double &y, const int &a, const int &b ) CONST_FUNCTION { 
	int result;
	asm(
	"fcomi	%2, %1\n"
	"cmovae	%4, %0\n"
	: "=r" (result) : "r" (x), "r" (y), "r" (a), "rm" (b) );
	return result;
}

template <>
inline uint ifless<double, double>( const double &x, const double &y, const uint &a, const uint &b ) CONST_FUNCTION { 
	uint result;
	asm(
	"fcomi	%2, %1\n"
	"cmovae	%4, %0\n"
	: "=r" (result) : "r" (x), "r" (y), "r" (a), "rm" (b) );
	return result;
}

template <>
inline double ifless<double, double>( const double &x, const double &y, const double &a, const double &b ) CONST_FUNCTION { 
	double result;
	asm(
	"fcomi	%2, %1\n"
	"fcmovnb	%4, %0\n"
	: "=r" (result) : "r" (x), "r" (y), "r" (a), "r" (b) );
	return result;
}

#endif

} // namespace mak

#endif // __maxmin_h
