// $Id: profile.cpp,v 1.6 2003/05/02 08:17:14 t Exp $ 
//
//    Copyright (c) 2003 Makino Takaki, t-libmak@snowelm.com
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#include "config_maklib.h"

#include <cctype>

#include <iostream>
#include <iomanip>
#if defined(LIBMAK_HAVE_HASH_MAP)
#include <hash_map>
#elif defined(LIBMAK_HAVE_EXT_HASH_MAP)
#include <ext/hash_map>
#endif

#include "profile.h"
//#include "hash_string.h"

namespace mak {

using namespace std;
#if defined(LIBMAK_HAVE_HASH_IN_GNU)
using __gnu_cxx::hash_map;
#elif defined(LIBMAK_HAVE_HASH_IN_STD)
using std::hash_map;
#endif

static const char rcsid[] = "$Id: profile.cpp,v 1.6 2003/05/02 08:17:14 t Exp $";

#include <math.h>

#ifdef MSDOS
#include <windows.h>
#else		// UNIX
#include "sys/time.h"
#include "sys/resource.h"
#endif

//#define DEBUG_PROF

timeunit GetTick() throw()
{
#if defined( HAVE_SYS_RESOURCE_H )
	struct rusage ru;
	getrusage(RUSAGE_SELF, &ru);

	return ru.ru_utime.tv_sec + ru.ru_utime.tv_usec * 0.000001;
#elif defined( HAVE_WINDOWS_H )

//	return GetTickCount() * 0.001;

	FILETIME c, e, k, u;
	GetProcessTimes(GetCurrentProcess(), &c, &e, &k, &u);
	
	return u.dwLowDateTime * 0.0000001 + u.dwHighDateTime * (0.0000001 * 4294967296.0);
#else
	return 0.0;
#endif
}


#ifdef LIBMAK_PROFILE

struct compare_profbase : public binary_function<profunit *, profunit *, bool>
{
	bool operator() (profunit * a, profunit * b);
};

bool compare_profbase::operator() (profunit * a, profunit * b) { return a->GetName() < b->GetName(); }


#ifdef LIBMAK_PASSIVE_PROFILE

#define TICK 10000		/* ticks; microsec */

const timeunit profiler_t::tickunit = TICK / 1000000.0;

#define USE_SIGACTION
#define USE_SIGALRM_HANDLER

#endif

// Global variables.

typedef hash_map<string, profunit *> profhash_t;
static profhash_t profhash;

ostream& profunit::output(ostream& os)
{
	profunit &p = *this;
//#ifdef LIBMAK_PASSIVE_PROFILE
//	timeunit alltime = p.alltime + p.exctime;
//	timeunit runtime = p.alltime;
//#else
	timeunit alltime = p.alltime;
	timeunit runtime = p.alltime - p.exctime;
	if( alltime < 0 )
		alltime = 0;
	if( runtime < 0 )
		runtime = 0;
//#endif
	timeunit pertime = runtime / p.passcnt;
//	os << "(" << alltime << "-" << p.exctime << "=" << runtime << ")" << endl;
	return os << setw(9) << p.passcnt
	          << setw(9) << (long)floor(alltime) << "." << setw(4) << setfill('0') << floor((alltime-floor(alltime))*10000) << setfill(' ')
	          << setw(9) << (long)floor(runtime) << "." << setw(4) << setfill('0') << floor((runtime-floor(runtime))*10000) << setfill(' ')
	          << setw(8) << (long)floor(pertime) << "." << setw(9) << setfill('0') << (long)(floor((pertime-floor(pertime))*1000000000)) << setfill(' ')
	          << "  " << p.name << endl;
}

profbase::profbase(const char *n)  throw()
{
#ifdef DEBUG_PROF
	cout << "PROF START: " << n << endl;
#endif
	parent = profiler.GetCurrent();
	profiler.SetCurrent(NULL);
	profhash_t::iterator i = profhash.find(string(n));
	if( i == profhash.end() )
	{
		my_unit = new profunit(n);
		profhash[n] = my_unit;
	}
	else
	{
		my_unit = i->second;
		my_unit->Enter();
	}
#ifdef LIBMAK_PASSIVE_PROFILE
	start_tick = 0;
#else
	start_tick = GetTick();
#endif
	profiler.SetCurrent(this);
}

profbase::~profbase() throw()
{
#ifdef DEBUG_PROF
	cout << "PROF END: " << my_unit->GetKey() << endl;
#endif
#ifdef LIBMAK_PASSIVE_PROFILE
	profiler.SetCurrent(NULL);
	if( my_unit->Leave() == 0 )
		my_unit->AddTime(start_tick);
	if( parent != NULL ) {
		parent->AddTime(start_tick);
		parent->my_unit->ExcTime(start_tick);
#ifdef DEBUG_PROF
		if( start_tick > 0.0 )
		cout << "PROF: " << parent->my_unit->GetKey() << " -> " << my_unit->GetKey() << ": " << start_tick << endl;
#endif
	}
#else
	timeunit end_tick = GetTick();
	timeunit tick = end_tick - start_tick;
	my_unit->AddTime(tick);
	if( parent != NULL ) {
		parent->my_unit->ExcTime(tick);
#ifdef DEBUG_PROF
		cout << "PROF: " << parent->my_unit->GetKey() << " -> " << my_unit->GetKey() << ": " << tick << endl;
#endif
	}
#endif
	profiler.SetCurrent(parent);
}

profiler_t::profiler_t()
{
	totalprof = new profbase("Total");

#ifdef LIBMAK_PASSIVE_PROFILE
//#if defined(USE_SIGACTION)
	struct sigaction act;
	
//# if defined(USE_SIGALRM_HANDLER)
	act.sa_handler = &SIGALRM_Handler;
	act.sa_flags = SA_RESTART;			// Not to stop FLEX action
	sigemptyset(&act.sa_mask);
	sigaction(SIGPROF, &act, &oldact_SIGALRM);
//# endif // USE_SIGALRM_HANDLER
//#endif // USE_SIGACTION

#ifdef USE_TIMER_SETTIME
	int ret = timer_create(CLOCK_REALTIME, NULL, &timerid);
	if( ret != 0 )
	{
		cout << "Fail to create timer" << endl;	return;
	}

	struct timespec tspec;

	clock_getres(CLOCK_REALTIME, &tspec);

	struct itimerspec itspec;

	itspec.it_interval.tv_sec  = 0;
	itspec.it_interval.tv_nsec = (TICK % 1000000) * 1000;
	itspec.it_value   .tv_sec  = 0;
	itspec.it_value   .tv_nsec = (TICK % 1000000) * 1000;

	cout << "Profile clock resolution:" << itspec.it_interval.tv_sec + itspec.it_interval.tv_nsec * 0.000000001 << endl;

	ret = timer_settime(timerid, 0, &itspec, NULL);

	if( ret != 0 )
	{
		cout << "Fail to set timer" << endl; return;
	}

	timer_gettime(timerid, &itspec);
	cout << "System clock resolution:" << itspec.it_interval.tv_sec + itspec.it_interval.tv_nsec * 0.000000001 << endl;
#else
	struct itimerval it;

	it.it_value.tv_sec = 0;
	it.it_value.tv_usec = TICK % 1000000;
	it.it_interval.tv_sec = 0;
	it.it_interval.tv_usec = TICK % 1000000;
	int ret = setitimer(ITIMER_PROF, &it, NULL);
	if( ret != 0 )
	{
		cout << "Fail to create timer" << endl;	return;
	}
#endif
#endif
}

profiler_t::~profiler_t()
{
	if( ! IsProfiling() )
		throw exception_prof("Profile consistency error!!");

#ifdef USE_TIMER_SETTIME
	timer_delete(timerid);
#endif

#ifdef LIBMAK_PASSIVE_PROFILE
	struct itimerval it;

	it.it_value.tv_sec = 0;
	it.it_value.tv_usec = 0;
	it.it_interval.tv_sec = 0;
	it.it_interval.tv_usec = 0;
	setitimer(ITIMER_PROF, &it, NULL);	// Stop timer

	sigaction(SIGPROF, &oldact_SIGALRM, NULL);
#endif
	
	delete totalprof;
#ifdef WIN95
	cout << "Profiling result (Win95 ticks; not accurate)" << endl;
#else
	cout << "Profiling result" << endl;
#endif
	cout << "PassCount    Total Time  Exclude Time   Time Per Pass    Name" << endl;

	vector<profunit *> arr;
	for( hash_map<string, profunit *>::const_iterator i = profhash.begin(); i != profhash.end(); i++ )
	{
		arr.push_back(i->second);
	}
	sort(arr.begin(), arr.end(), compare_profbase());
	for( vector<profunit *>::iterator i=arr.begin(); i != arr.end(); i++ )
	{
		(*i)->output(cout);
		delete *i;
	}
}

void profiler_t::EnableProfiling() throw(exception_prof)
{ 
	if( profile_block_count <= 0 ) 
		throw exception_prof("Bad Profiling"); 
	profile_block_count--; 
}

profiler_t profiler;
/*
timeunit tm = 0;
long timecnt = 0;
long overrun = 0;
*/
#ifdef LIBMAK_PASSIVE_PROFILE
void profiler_t::SIGALRM_Handler(int)
{
//	long ov = timer_getoverrun(profr->timerid);
	if( profiler.IsProfiling() )
	{
		profbase *p = profiler.current;
		if( p != NULL )
		{
			p->AddTime(tickunit);// * (1 + ov));
		}
	}
/*
	else
	{
		tm += tickunit;// * (1 + ov);
	}
*/
//	overrun += ov;
}
#endif

} // namespace mak

#endif
