//
///  <ID>$Id: ref.h,v 1.4 2003/04/28 05:49:15 t Exp $</ID>
//
///  <COLLECTION>MakLib Code Library</COLLECTION>
//
///  <name>ref.h</name>
//
///  <overview>
///  <jpn>QƃJE^t|C^^</jpn>
///  <eng>Pointer class with reference counter. </eng>
///  </overview>
//
//    Copyright (c) 2003 Makino Takaki, t-libmak@snowelm.com
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#ifndef __ref_h
#define __ref_h

namespace mak {

/////////////////////////////////////////////////////////////////////////

/// <classdef>
/// <name> ref_base </name>

/// <overview>ǗNX</overview>
/// <desc>
/// <jpn>
/// <p>
/// <ref>ref</ref> NX̃x[XNXłB
/// ̗Rɂ葶݂Ă܂Ǎ^𒼐ڃvOŎgp邱Ƃ͂܂B
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// not yet
/// </p>
/// </eng>
/// </desc>

/// <body>

class ref_base
{
protected:
	const void *p;				/// pointer to the object
	unsigned int *cnt;		/// pointer to the counter 

	ref_base( const void *ip, unsigned int *icnt ) : p(ip), cnt(icnt) { }
							/// constructor is internal
	template <class T, class U>
	static const void *convert( const ref_base &b ) { return static_cast<const void *>( static_cast<const T *>( static_cast<const U *>( b.p ) ) ); }

	static unsigned int *getcntptr( const ref_base &b ) { return b.cnt; }
}; 

/// </body>
/// </classdef>

/////////////////////////////////////////////////////////////////////////

/// <classdef>
/// <name> const_ref </name>

/// <overview><jpn>QƃJE^t|C^NX</jpn><eng>Pointer class with reference counter</eng></overview>
/// <desc>
/// <jpn>
/// <p>
/// QƃJE^|C^NXłB
/// ׂĂ const_ref  destruct ꂽ_ŎQƐ̃IuWFNg delete ܂B
/// ̃IuWFNgŵȂ̂ŁA|C^ɑ΂čs鑀̂zANZXɊւ (Ƃ ++  --) Ƃ͂ł܂B
/// ȊO̓_ƂĂ͒ʏ̃|C^Ɠ悤Ɉ܂B
/// </p>
/// <p>
/// Ŝ߁ARs[RXgN^ȊO͈Öق̕ϊł܂B
/// ܂AC `̃|C^ɖ߂Ƃł܂B
/// </p>
/// <p>
/// ref ^̃|C^Vmۂďɂ <ref>New</ref> ֐gĂB
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// not yet
/// </p>
/// </eng>
/// </desc>
/// <see> <ref>New()</ref>, <ref>Ref</ref>, ref </see>

/// <body>

template <class T>
class const_ref : public ref_base
{
public:
	typedef T referer_t;
			/// Type of the referer.
protected:
	const referer_t *getptr() const { return static_cast<const T *>(p); }
	void setptr(const referer_t *ptr) { p = static_cast<const void *>(ptr);}

public:
	const_ref() : ref_base(NULL, NULL) { }
				/// The default constructor makes a NULL pointer.
	explicit const_ref(const referer_t *t) : ref_base( static_cast<const void *>(t), new unsigned int(1) ) { }
				/// Makes a pointer with reference count 1.
				/// You should use <ref>New()</ref> instead of using this constructor.
	explicit const_ref(const referer_t &t) : ref_base( static_cast<const void *>(&t), NULL ) { }
				/// Makes a pointer with no reference count (never deleted)
				/// You should use <ref>Ref()</ref> instead of using this constructor.

	const_ref(const const_ref<T> &p1) : ref_base( p1.p, p1.cnt ) { if(cnt) (*cnt)++; }
				/// Copy constructor increments the reference count by 1.
	template <class U> const_ref(const const_ref<U> &u) 
		: ref_base( convert<T,U>(u), getcntptr(u) ) { if(cnt) (*cnt)++; }
				/// Constructor from compatible pointer class (upcast only).

	~const_ref() { if( cnt && --(*cnt) == 0 ) { delete cnt; delete getptr(); } }
				/// Destructor decrements the reference count by 1, and deletes the object if the count reaches to 0.

	const_ref<T> &operator =(const const_ref<T> &p1) 
				/// Assignment function first decrements the reference count of the left hand object, and delete it if the count reaches to 0.
				/// After that, assignment is done and the counter is incremented by 1.
		{ if( cnt && --(*cnt) == 0 ) 
			{ delete cnt; delete getptr(); } 
		p = p1.p; cnt = p1.cnt; 
		if(cnt) (*cnt)++; 
		return *this; }

	const referer_t * operator ->() const { return getptr(); }

	const referer_t & operator *() const { return *getptr(); }

	friend bool operator==(const const_ref<T> &r1, const const_ref<T> &r2) { return r1.p == r2.p; }
				/// Comparison of two pointers.
	friend bool operator!=(const const_ref<T> &r1, const const_ref<T> &r2) { return r1.p != r2.p; }

	friend bool operator==(const const_ref<T> &r, void *p) { return r.p == p; }
	friend bool operator==(void *p, const const_ref<T> &r) { return r.p == p; }
				/// These operators are intended to compare the ref object and a null pointer.
	friend bool operator!=(const const_ref<T> &r, void *p) { return r.p != p; }
	friend bool operator!=(void *p, const const_ref<T> &r) { return r.p != p; }
				/// These operators are intended to compare the ref object and a null pointer.
};
/// </body>
/// </classdef>

/////////////////////////////////////////////////////////////////////////

/// <classdef>
/// <name> ref </name>

/// <overview><jpn>QƃJE^t|C^NX</jpn><eng>Pointer class with reference counter</eng></overview>
/// <desc>
/// <jpn>
/// <p>
/// QƃJE^|C^NXłB
/// ׂĂ ref  destruct ꂽ_ŎQƐ̃IuWFNg delete ܂B
/// ̃IuWFNgŵȂ̂ŁA|C^ɑ΂čs鑀̂zANZXɊւ (Ƃ ++  --) Ƃ͂ł܂B
/// ȊO̓_ƂĂ͒ʏ̃|C^Ɠ悤Ɉ܂B
/// </p>
/// <p>
/// Ŝ߁ARs[RXgN^ȊO͈Öق̕ϊł܂B
/// ܂AC `̃|C^ɖ߂Ƃł܂B
/// </p>
/// <p>
/// ref ^̃|C^Vmۂďɂ <ref>New</ref> ֐gĂB
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// not yet
/// </p>
/// </eng>
/// </desc>
/// <see> <ref>New()</ref>, <ref>Ref</ref>, const_ref </see>

/// <body>

template <class T>
class ref : public const_ref<T>
{
public:
	typedef T referer_t;
			/// Type of the referer.
protected:
	referer_t *getptr() const { return const_cast<T *>(const_ref<T>::getptr()); }

public:
	ref() : const_ref<T>() { }
				/// The default constructor makes a NULL pointer.
	explicit ref(referer_t *t) : const_ref<T>(t) { }
				/// Makes a pointer with reference count 1.
				/// You should use <ref>New()</ref> instead of using this constructor.
	explicit ref(referer_t &t) : const_ref<T>(t) { }
				/// Makes a pointer with no reference count (never deleted)
				/// You should use <ref>Ref()</ref> instead of using this constructor.

	ref(const ref<T> &p1) : const_ref<T>(p1) {  }
				/// Copy constructor increments the reference count by 1.
	template <class U> ref(const ref<U> &u) 
		: const_ref<T>(u) {  }
				/// Constructor from compatible pointer class (upcast only).

	ref<T> &operator =(const ref<T> &p1) 
				/// Assignment function first decrements the reference count of the left hand object, and delete it if the count reaches to 0.
				/// After that, assignment is done and the counter is incremented by 1.
		{ if( cnt && --(*cnt) == 0 ) 
			{ delete cnt; delete getptr(); } 
		p = p1.p; cnt = p1.cnt; 
		if(cnt) (*cnt)++; 
		return *this; }

	referer_t * operator ->() const { return getptr(); }
				/// operator -> is the way to access the member of the object.

	referer_t & operator *() const { return *getptr(); }
				/// operator * is the way to access the object itself.
};
/// </body>
/// </classdef>


/// <funcdef>
/// <name> New </name>
/// <overview><jpn><ref>ref</ref>&lt;T&gt; ̍\z֐</jpn><eng>Construction function of <ref>ref</ref>&lt;T&gt;</eng></overview>
/// <decl> 
/// #include &lt;<ref>ref.h</ref>&gt;<br><br>
/// template &lt;class T&gt; <br> ref&lt;T&gt; New( ... ); 
/// </decl>
/// <desc>
/// <p><code>new T(...)</code>  <ref>ref</ref>&lt;T&gt; ^Ԃo[WłB</p>
/// <p><code>new T(...)</code> ̂ <code>New&lt;T&gt;(...)</code> ƏĂBV̈mۂāAT ^̃IuWFNgzuA<ref>ref</ref>&lt;T&gt; ^Ԃ܂B</p>
/// <samp> some_class * @ some_ptr = new some_class ("Constructor Arguments", 123, xxx);</samp> <br>@@@// ʏ̃|C^̊m (Ȃ) <br>
/// <samp> ref&lt;some_class&gt; some_ptr = New&lt;some_class&gt;("Constructor Arguments", 123, xxx);</samp> <br>@@@// ref |C^̊m ()
/// </desc>
/// <args>
/// T ̃RXgN^ɓn̂܂܏܂B
/// </args>
/// <retval>
/// VmۂꂽIuWFNgw <ref>ref</ref><T> ^̃IuWFNgłB
/// </retval>
/// <remark> 
/// <p> RXgN^ɔ const ̈nꍇɃG[o邱Ƃ܂B</p>
/// <p> RXgN^̈ 6 ȏ゠ꍇɂ <ref>ref.h</ref> Kv܂B</p></remark>
/// <see> <ref>ref</ref>, <ref>Ref</ref> </see>
/// <body>

template <class T> inline
ref<T> New( ) { return ref<T>(new T()); }
template <class T, class T1> inline
ref<T> New( const T1 &t1 ) { return ref<T>(new T(t1)); }
template <class T, class T1, class T2> inline
ref<T> New( const T1 &t1, const T2 &t2 ) { return ref<T>(new T(t1, t2)); }
template <class T, class T1, class T2, class T3> inline
ref<T> New( const T1 &t1, const T2 &t2, const T3 &t3 ) { return ref<T>(new T(t1, t2, t3)); }
template <class T, class T1, class T2, class T3, class T4> inline
ref<T> New( const T1 &t1, const T2 &t2, const T3 &t3, const T4 &t4 ) { return ref<T>(new T(t1, t2, t3, t4)); }
template <class T, class T1, class T2, class T3, class T4, class T5> inline
ref<T> New( const T1 &t1, const T2 &t2, const T3 &t3, const T4 &t4, const T5 &t5 ) { return ref<T>(new T(t1, t2, t3, t4, t5)); }
template <class T, class T1, class T2, class T3, class T4, class T5, class T6> inline
ref<T> New( const T1 &t1, const T2 &t2, const T3 &t3, const T4 &t4, const T5 &t5, const T6 &t6 ) { return ref<T>(new T(t1, t2, t3, t4, t5, t6)); }

/// </body>
/// </funcdef>

/// <funcdef>
/// <name> Ref </name>
/// <overview><jpn><ref>ref</ref>&lt;T&gt; ̍\z֐</jpn><eng>Construction function of <ref>ref</ref>&lt;T&gt;</eng></overview>
/// <desc>
/// 
/// <p><ref>ref</ref>&lt;T&gt; ̃IuWFNg̎QƂ邽߂̊֐łB</p>
/// <p>some_class ̕ϐ some_var ƂARef(some_var)  some_var wiJE^̂Ȃjref QƂ
/// Ԃ܂B ref&lt;some_class&gt;(some_var) ƓłAsome_var ̌^LqȂĂ悭ȂĂ܂B</p>
/// </desc>
/// <args>
/// <p>ɂȂIuWFNg̎QƂn܂B|C^\z͂łȂ (|C^ւ̎QƂɂȂ) ̂ŒӂĂB</p>
/// </args>
/// <retval>
/// <p>w肳ꂽIuWFNgw <ref>ref</ref><T> ^̃IuWFNgłB</p>
/// </retval>
/// <remark> 
/// <p> Zq new ŊmۂIuWFNgɑ΂ Ref Ńt@XIuWFNg쐬ĂÃIuWFNg͍̎s܂BKvȏꍇ <ref>New</ref> gĂB</p>
/// </remark>
/// <see> <ref>ref</ref>, <ref>New</ref> </see>
/// <body>

template <class T> inline
const_ref<T> Ref(const T &t) { return const_ref<T>(const_cast<T &>(t)); }
template <class T> inline
ref<T> Ref(T &t) { return ref<T>(t); }

/// </body>
/// </funcdef>

} // namespace mak

#endif // __ref_h
