//
///  <ID>$Id: stringop.h,v 1.3 2003/04/28 07:11:55 t Exp $</ID>
//
///  <COLLECTION>MakLib Code Library</COLLECTION>
//
///  <name>stringop.h</name>
//
///  <overview>
///  <jpn>string ̑֐</jpn>
///  <eng>Manipulation functions of string </eng>
///  </overview>
//
//    Copyright (c) 2003 Makino Takaki, t-libmak@snowelm.com
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#ifndef __stringop_h
#define __stringop_h

#include "config_maklib.h"

#include <iostream>
#include <string>
#ifdef LIBMAK_HAVE_SSTREAM
#include <sstream>
#else
#include <strstream>
#endif

namespace mak {

/////////////////////////////////////////////////////////////////////////

/// <funcdef>
/// <name> makestr </name>

/// <overview><jpn>string ɕϊ</jpn><eng>Convert to string</eng></overview>
/// <desc>
/// <jpn>
/// <p>
/// ostream ւ &lt;&lt; Ɠlȕ@ŁAl𕶎ɕϊ܂B
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// Convert a value to string with the similar way of operator &lt;&lt;(ostream &, ...).
/// </p>
/// </eng>
/// </desc>

/// <body>

template <class T>
std::string makestr(const T &t)
{
#ifdef LIBMAK_HAVE_SSTREAM
	std::ostringstream oss;
	oss << t;
	return oss.str();
#else
	std::ostrstream oss;
	oss << t << std::ends;
	std::string ret( oss.str() );
	delete[] oss.str();
	return ret;
#endif
}
/// </body></funcdef>


/////////////////////////////////////////////////////////////////////////

/// <funcdef>
/// <name> quotestr </name>

/// <overview><jpn>string  " </jpn><eng>Put quotation to the string</eng></overview>
/// <desc>
/// <jpn>
/// <p>
///  "" ɂ܂B "  \ ꍇ́Aꂼ \"  \\ ɕϊ܂B
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// Put double quotation marks for a string (that is, "string"). 
/// " and \ are converted to \" and \\, respectively.
/// </p>
/// </eng>
/// </desc>

/// <body>

inline std::string quotestr(std::string s)
{
	for( std::string::size_type p=0; p < s.size(); p++ )
	{
		if( s[p] == '\"' || s[p] == '\\' )
		{
			s.insert(p, "\\");
			p++;
		}
	}
	return "\"" + s + "\"";
}

/// </body></funcdef>

} // namespace mak

#endif // __stringop_h
