///  <ID>$Id: test.h,v 1.4 2003/04/28 05:49:15 t Exp $</ID>
//
///  <COLLECTION> neurein </COLLECTION>
//
///  <name>cppUnit.h</name>
//
///  <overview>
///  <jpn>eXg/AT[VpNX</jpn>
///  <eng>unit testing routines</eng>
///  </overview>
//
//    Copyright (c) 2003 Makino Takaki, t-libmak@snowelm.com
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#ifndef __test_h
#define __test_h

#include <iostream>
#include <sstream>
#include <set>
#include <string>
#include <typeinfo>
#include <exception>

#include <cassert>
#if defined(assert)
#undef assert		// Disable original assertion
#endif

#define assert(x)		( (x) ? (void)0 : mak::_assert(#x, __FILE__, __LINE__) )

#define testEqual(x,y)		_testEqual(x, y, #x, #y, __FILE__, __LINE__)
#define testGreater(x,y)		_testGreater(x, y, #x, #y, __FILE__, __LINE__)
#define testLess(x,y)		_testLess(x, y, #x, #y, __FILE__, __LINE__)
#define testGreaterEqual(x,y)		_testGreaterEqual(x, y, #x, #y, __FILE__, __LINE__)
#define testLessEqual(x,y)		_testLessEqual(x, y, #x, #y, __FILE__, __LINE__)
#define testRange(x,y,z)		_testRange(x, y, z, #x, #y, #z, __FILE__, __LINE__)
#define testEqualDelta(x,y,delta)	_testEqualDelta(x, y, delta, #x, #y, __FILE__, __LINE__)
#define testThrow(command, excep)		do{ try { { command; } _testFailThrow(#command, #excep, __FILE__, __LINE__); } \
										catch(excep) { _testSuccessThrow(#command, #excep, __FILE__, __LINE__); }  \
										catch(std::exception &e) { _testFailThrow(#command, #excep, e, __FILE__, __LINE__); } \
										catch(...) { _testFailThrowX(#command, #excep, __FILE__, __LINE__); } } while(0)
										
#define testNoThrow(command)		do { try { { command; } _testSuccessNoThrow(#command, __FILE__, __LINE__); } \
										catch(std::exception &e) { _testFailNoThrow(#command, e, __FILE__, __LINE__); } \
										catch(...) { _testFailNoThrow(#command, __FILE__, __LINE__); } } while(0)
#define testTrue(x)		testEqual(x, true)
#define testFalse(x)		testEqual(x, false)

namespace mak {

class TestSuite;

/// <classdef>
/// <name> TestBase </name>
/// <overview>܂܂ȃeXg`邽߂̏zNX</overview>
/// <desc>
/// <p>
/// jbgeXg`邽߂̊NXłB̃NX[U[ڎgƂ͂܂肠܂B
/// ʏ́ATestATestBundle ̂ꂩ̃NXpėp邩ATestSuite NX̃CX^X܂B
/// </p>
/// </desc>
/// <see>Test, TestBundle, TestSuite</see>
/// <body>

class TestBase
{
protected:
	TestSuite *suite;
		/// ̃eXgeXgXC[g\܂B

	template <class T> 
	void _testEqual( const T &a, const T &b, const char *astr, const char *bstr, const char *file, int line );
		/// testEqual(x,y)  x  y ƂeXg܂B
	template <class T> 
	void _testGreater( const T &a, const T &b, const char *astr, const char *bstr, const char *file, int line );
		/// testGreater(x,y)  x > y eXg܂B
	template <class T> 
	void _testLess( const T &a, const T &b, const char *astr, const char *bstr, const char *file, int line );
		/// testLess(x,y)  x < y eXg܂B
	template <class T> 
	void _testGreaterEqual( const T &a, const T &b, const char *astr, const char *bstr, const char *file, int line );
		/// testEqual(x,y)  x >= y eXg܂B
	template <class T> 
	void _testLessEqual( const T &a, const T &b, const char *astr, const char *bstr, const char *file, int line );
		/// testEqual(x,y)  x <= y eXg܂B
	template <class T> 
	void _testRange( const T &a, const T &b, const T &c, const char *astr, const char *bstr, const char *cstr, const char *file, int line );
		/// testrange(x,y,z)  x <= y <= zeXg܂B

	void _testEqual( std::string a, std::string b, const char *astr, const char *bstr, const char *file, int line );

	template <class T> 
	void _testEqualDelta( const T &a, const T &b, const T &delta, const char *astr, const char *bstr, const char *file, int line );
		/// testEqualDelta(x,y,delta)  x  y  delta ȓœǂeXg܂B

//	void _testEqual( int a, int b, const char *astr, const char *bstr, const char *file, int line );
//	void _testGreater( int a, int b, const char *astr, const char *bstr, const char *file, int line );
//	void _testLess( int a, int b, const char *astr, const char *bstr, const char *file, int line );
//	void _testGreaterEqual( int a, int b, const char *astr, const char *bstr, const char *file, int line );
//	void _testLessEqual( int a, int b, const char *astr, const char *bstr, const char *file, int line );
//	void _testRange( int a, int b, int c, const char *astr, const char *bstr, const char *cstr, const char *file, int line );
//	void _testEqualDelta( int a, int b, int delta, const char *astr, const char *bstr, const char *file, int line );


	void _testFailThrow( const char *astr, const char *bstr, const char *file, int line );
	void _testFailThrow( const char *astr, const char *bstr, std::exception &excep, const char *file, int line );
	void _testFailThrowX( const char *astr, const char *bstr, const char *file, int line );
	void _testSuccessThrow( const char *astr, const char *bstr, const char *file, int line );
		/// testThrow(x,y) ́A x O y 𑗏o邱ƂeXg܂B
	void _testFailNoThrow( const char *astr, const char *file, int line );
	void _testFailNoThrow( const char *astr, std::exception &excep, const char *file, int line );
	void _testSuccessNoThrow( const char *astr, const char *file, int line );
		/// testNoThrow(x) ́A x O𑗏oȂƂeXg܂B

public:
	TestBase(TestSuite *is = NULL);
		/// eXgXC[g is ɑeXg쐬܂BNULL w肵ꍇ́AeXgXC[gɑȂeXgɂȂ܂B
	virtual ~TestBase();
		/// eXg̃fRXgN^łBeXgXC[gɑĂꍇ́ÃeXgXC[g玩폜܂B

	virtual const char *name() const { return typeid(*this).name(); }
		/// ̃eXgXC[g̃NXԂ܂B

	virtual void success(std::string desc = "Success") throw(); 
		/// eXgꍇ test() ĂяoR[obNłB
	virtual void failure(std::string desc = "Failure") throw(); 
		/// eXgsꍇ test() ĂяoR[obNłB
	virtual void failureExpect(bool expect=true) throw();
		/// ̃eXgs邱Ƃ\Ăꍇ test() ĂяoR[obNłB

	virtual void runTest() = 0;
		/// eXgs܂B

private:
	friend class TestSuite;
	void setSuite(TestSuite *p) { suite = p; }
		/// TestSuite NXƋ邽߂̊֐łB
	TestSuite *getSuite() { return suite; }
		/// TestSuite NXƋ邽߂̊֐łB
};
/// </body></classdef>

/// <classdef>
/// <name> Test </name>
/// <overview>eXg`邽߂̏zNX</overview>
/// <desc>
/// <p>
/// jbgeXg`邽߂̊NXłB
/// </p>
/// </desc>
/// <see>TestBundle, TestSuite</see>
/// <body>
class Test : public TestBase
{
public:
	Test(TestSuite *is = NULL) : TestBase(is) { }
		/// eXgXC[g is ɑeXg쐬܂BNULL w肵ꍇ́AeXgXC[gɑȂeXgɂȂ܂B
	virtual void runTest();
		/// eXgs܂B

protected:
	virtual void setUp() { }
		/// eXgO֐łBhNXɂĒu邽߂Ɏg܂B
	virtual void test() = 0;
		/// eXg{̂łBhNXɂĒu܂B
	virtual void tearDown() { }
		/// eXg㏈֐łBhNXɂĒu邽߂Ɏg܂B
};
/// </body></classdef>

/// <classdef>
/// <name> TestBundle </name>
/// <overview>̒PeXg̉`邽߂̏zNX</overview>
/// <desc>
/// <p>
/// jbgeXg`邽߂̊NXłB
/// </p>
/// </desc>
/// <see>Test, TestSuite</see>
/// <body>
class TestBundle : public TestBase
{
public:
	TestBundle(TestSuite *is = NULL) : TestBase(is) { }

	virtual void runTest();
		/// eXgs܂B
	virtual void runTest(int testno);
		/// w肳ꂽeXgs܂B

protected:
	virtual int testCount() = 0;
	
	virtual void setUp() { }
		/// eXgO֐łBhNXɂĒu邽߂Ɏg܂B
	virtual void test(int i) = 0;
		/// eXg{̂łBhNXɂĒu܂B
	virtual void tearDown() { }
		/// eXg㏈֐łBhNXɂĒu邽߂Ɏg܂B
};
/// </body></classdef>

/// <classdef>
/// <name> TestSuite </name>
/// <overview>eXg̏W`邽߂̏zNX</overview>
/// <desc>
/// <p>
/// eXgW\NXłBeXg̃RXgN^Ɏw肷邱ƂŁA
/// IɃeXgɑ`ɂȂ܂B
/// </p>
/// </desc>
/// <see>Test</see>
/// <body>
class TestSuite : public TestBase
{
protected:
	typedef std::set<TestBase *> tests_type;
		/// ̃eXgXC[gɑeXgi[ <var>tests</var> ̌^łB

	static TestSuite *root;
		/// [g`FC̊Jnn_łB
	TestSuite *next;
		/// [g`FC̎̃XC[głB
		/// Ō̃XC[g̏ꍇA[g`FCɑȂꍇ 
		/// (<var>suite</var> oϐ NULL łȂꍇ)  NULL ɂȂ܂B

	std::string name;
		/// ̃eXgXC[g̖OłB
	std::ostream *os;
		/// ̃eXgXC[ǧʂ̏o͐łB
	tests_type tests;
		/// ̃eXgXC[gɑeXg̏WłB
	int succcnt;
		/// eXgXC[g̃eXg̐񐔂łB
	int failcnt;
		/// eXgXC[g̃eXg̎s񐔂łB
	int unexcnt;
		/// eXgXC[g̃eXg̗\Ȃs̉񐔂łB
	int verbose;
		/// eXgXC[g̕\ʂłB0,1,2 Ӗ̂lłB
	bool failureExpected;
		/// ̃eXg̎s\Ă邩ǂ܂B

public:
	TestSuite(std::string iname);
		/// [g`FCɑAO <var>iname</var> ̃eXgXC[g쐬܂B
	TestSuite(std::string iname, TestSuite *parent);
		/// eeXgXC[g <var>parent</var> ɑAO <var>iname</var> ̃eXgXC[g쐬܂B
	virtual ~TestSuite();
		/// eXgXC[g̃fRXgN^łB[g`FCɑꍇ
		/// [g`FC玩g폜܂B܂A܂܂eXg̑eXgXC[g
		/// ݒ܂B
	
	void Add(TestBase *t);
		/// ̃eXgXC[gɎw肳ꂽeXg <var>t</var> ǉ܂B<var>t</var> 
		/// łɕʂ̃eXgXC[gɑĂꍇ́ÃeXgXC[gO܂B
	void Remove(TestBase *t);
		/// ̃eXgXC[g炵ĂꂽeXg <var>t</var> 폜܂B<var>t</var> 
		/// ܂܂ĂȂꍇ͉܂B

	virtual void setUp();
		/// eXgXC[g̃eXgO܂B/sJEg̃NA܂B
	virtual void tearDown();
		/// eXgXC[g̃eXg㏈܂BeeXgXC[gȂAverbose  1 ȏ̏ꍇA
		/// ̃eXǧʗv os ɏo͂܂B

	virtual void success(std::string desc = "Success") throw(); 
		/// eXgXC[gɑeXgꍇɌĂяo܂B
	virtual void failure(std::string desc = "Failure") throw(); 
		/// eXgXC[gɑeXgꍇɌĂяo܂B
	virtual void failureExpect(bool expect=true) throw();
		/// eXgXC[gŎ̃eXgs邱Ƃ\ĂꍇɌĂяo܂B

	virtual void runTest();

	virtual std::ostream &stream() const { return *os; }
		/// ݂̏o͐Xg[Ԃ܂B
	virtual void setStream(std::ostream &o) { os = &o; }
		/// o͐Xg[ݒ肵܂B
	void setVerbose(int v) { verbose = v; }
		/// bZ[Wʂݒ肵܂B0,1,2 ̂ꂩLȒlłB
	int getSuccessCount() const { return succcnt; }
		/// eXg̐񐔂擾܂B
	int getFailureCount() const { return failcnt; }
		/// eXg̎s񐔂擾܂B
	int getUnexpectedFailureCount() const { return unexcnt; }
		/// eXg̎s񐔂擾܂B

	static void setStreamAll(std::ostream &o);
	static void setVerboseAll(int v);
	static int testAll();
	static int getSuccessCountAll();
	static int getFailureCountAll();
};
/// </body></classdef>

/////////////////////////////////////////////////////////////////////////

template <class T> inline
void TestBase::_testEqual( const T &a, const T &b, const char *astr, const char *bstr, const char *file, int line )
{
	std::ostringstream st;
	st << file << ":" << line << ": testEqual " << astr << " (" << a << ") == " << bstr << " (" << b << ")";
	if( a == b ) {
		success(st.str());
	}
	else {
		failure(st.str());
	}
}
template <class T> inline
void TestBase::_testGreater( const T &a, const T &b, const char *astr, const char *bstr, const char *file, int line )
{
	std::ostringstream st;
	st << file << ":" << line << ": testEqual " << astr << " (" << a << ") > " << bstr << " (" << b << ")";
	if( a > b ) {
		success(st.str());
	}
	else {
		failure(st.str());
	}
}
template <class T> inline
void TestBase::_testLess( const T &a, const T &b, const char *astr, const char *bstr, const char *file, int line )
{
	std::ostringstream st;
	st << file << ":" << line << ": testEqual " << astr << " (" << a << ") < " << bstr << " (" << b << ")";
	if( a < b ) {
		success(st.str());
	}
	else {
		failure(st.str());
	}
}
template <class T> inline
void TestBase::_testGreaterEqual( const T &a, const T &b, const char *astr, const char *bstr, const char *file, int line )
{
	std::ostringstream st;
	st << file << ":" << line << ": testEqual " << astr << " (" << a << ") > " << bstr << " (" << b << ")";
	if( a >= b ) {
		success(st.str());
	}
	else {
		failure(st.str());
	}
}
template <class T> inline
void TestBase::_testLessEqual( const T &a, const T &b, const char *astr, const char *bstr, const char *file, int line )
{
	std::ostringstream st;
	st << file << ":" << line << ": testEqual " << astr << " (" << a << ") < " << bstr << " (" << b << ")";
	if( a <= b ) {
		success(st.str());
	}
	else {
		failure(st.str());
	}
}
template <class T> inline
void TestBase::_testRange( const T &a, const T &b, const T &c, const char *astr, const char *bstr, const char *cstr, const char *file, int line )
{
	std::ostringstream st;
	st << file << ":" << line << ": testRange " << astr << " (" << a << ") <= " << bstr << " (" << b << ") <= " << cstr << " (" << c << ")";
	if( a <= b && b <= c ) {
		success(st.str());
	}
	else {
		failure(st.str());
	}
}
template <class T> inline
void TestBase::_testEqualDelta( const T &a, const T &b, const T &delta, const char *astr, const char *bstr, const char *file, int line )
{
	std::ostringstream st;
	st << file << ":" << line << ": testEqualDelta " << astr << " (" << a << ") == " << bstr << " (" << b << ") +/- " << delta;
	if( a - b > -delta && a - b < delta ) {
		success(st.str());
	}
	else {
		failure(st.str());
	}
}

/// <classdef>
/// <name> AssertFailedException </name>
/// <overview>AT[V̎s񍐂ONX</overview>
/// <desc>
/// <p>
/// ONXłBassert sƂ񍐂܂B
/// </p>
/// </desc>
/// <see>Test</see>
/// <body>
class AssertFailedException : public std::exception
{
	const char *s;
public:
	AssertFailedException(const char *teststr, const char *file, int line);
	virtual ~AssertFailedException() throw();
	virtual const char *what();
};
/// </body></classdef>

void assertFail(const char *teststr, const char *file, int line) throw(AssertFailedException);

inline void _assert(bool test, const char *teststr, const char *file, int line) throw(AssertFailedException)
{
	if( !test )
		assertFail(teststr, file, line);
}


}	// namespace mak

#endif
