//
///  <ID>$Id: typestr.h,v 1.2 2003/04/28 05:49:15 t Exp $</ID>
//
///  <COLLECTION>MakLib Code Library</COLLECTION>
//
///  <name>typestr.h</name>
//
///  <overview>
///  <jpn>^𕶎ƂĎ擾</jpn>
///  <eng>Obtain type name string, more reasonable than typeid</eng>
///  </overview>
//
//    Copyright (c) 2003 Makino Takaki, t-libmak@snowelm.com
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#ifndef __typestr_h
#define __typestr_h

#include <typeinfo>
#include <string>

namespace mak {

/////////////////////////////////////////////////////////////////////////

/// <classdef>
/// <name> typestring_traits </name>

/// <overview>^擾֐̂߂̃ev[gNX</overview>
/// <desc>
/// <jpn>
/// <p>
/// <ref>typestring</ref> Ŏgev[gNXłB
/// [U[ꉻo[W`邱ƂŁǍ^̖OI[o[Ch邱Ƃł܂B
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// A template class used in <ref>typestring</ref>.
/// You can override the name of any type by defining a user-specialized version of this class.
/// </p>
/// </eng>
/// <samp>
/// template &lt;&gt; struct typestring_traits&lt;class USA&gt; <br>
/// { static string get_typestring() { return "United States of America"; } };</samp>
/// </samp>
/// </desc>
/// <see>typestring</see>

/// <body>

template <class T> struct typestring_traits 
{ static std::string get_typestring() { return typeid(T).name(); } 
	/// <jpn>^ T ̌^\擾܂B</jpn><eng>Obtain a string that represents the type T.</eng>
};

/// </body></classdef>


/////////////////////////////////////////////////////////////////////////

/// <funcdef>
/// <name> typestring </name>

/// <overview>^擾֐</overview>
/// <desc>
/// <jpn>
/// <p>
/// ev[g T ̌^擾܂B
/// gݍ݌^ȊǑ^̏ꍇAftHg (typeinfo ɂ) ^擾܂B
/// 擾^ςꍇ́Ǎ^ɑ΂ <ref>typestring_traits</ref> ̃[U[ꉻo[W`ĂB
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// not yet
/// </p>
/// </eng>
/// </desc>
/// <see>typestring_traits</see>

/// <body>

template <class T> std::string typestring() { return typestring_traits<T>::get_typestring(); }

/// </body></funcdef>


template <> struct typestring_traits<char> 
{ static std::string get_typestring() { return "char"; } };

template <> struct typestring_traits<unsigned char> 
{ static std::string get_typestring() { return "unsigned char"; } };

template <> struct typestring_traits<signed char> 
{ static std::string get_typestring() { return "signed char"; } };

template <> struct typestring_traits<short> 
{ static std::string get_typestring() { return "short"; } };

template <> struct typestring_traits<unsigned short> 
{ static std::string get_typestring() { return "unsigned short"; } };

template <> struct typestring_traits<int> 
{ static std::string get_typestring() { return "int"; } };

template <> struct typestring_traits<unsigned int> 
{ static std::string get_typestring() { return "unsigned int"; } };

template <> struct typestring_traits<long> 
{ static std::string get_typestring() { return "long"; } };

template <> struct typestring_traits<unsigned long> 
{ static std::string get_typestring() { return "unsigned long"; } };

template <> struct typestring_traits<float> 
{ static std::string get_typestring() { return "float"; } };

template <> struct typestring_traits<double> 
{ static std::string get_typestring() { return "double"; } };

template <> struct typestring_traits<void> 
{ static std::string get_typestring() { return "void"; } };

template <> struct typestring_traits<std::string> 
{ static std::string get_typestring() { return "string"; } };

template <> struct typestring_traits<bool> 
{ static std::string get_typestring() { return "bool"; } };


template <class T, int I> struct typestring_traits<T[I]> 
{ static std::string get_typestring() { return typestring<T>() + " [" + inttostr(i) + "]"; } };

template <class T> struct typestring_traits<T *> 
{ static std::string get_typestring() { return typestring<T>() + " *"; } };

template <class T> struct typestring_traits<T &> 
{ static std::string get_typestring() { return typestring<T>() + " &"; } };

template <class T> struct typestring_traits<const T> 
{ static std::string get_typestring() { return "const " + typestring<T>(); } };

template <class T> struct typestring_traits<volatile T> 
{ static std::string get_typestring() { return "volatile " + typestring<T>(); } };

template <class T> struct typestring_traits<T()> 
{ static std::string get_typestring() { return typestring<T>() + "()"; } };

/// </body></classdef>

//template <class T> struct typestring_traits< set<T> > 
//{ static string get_typestring() { return "set< " + typestring<T>() + " >"; } };

//template <class T1, class T2> struct typestring_traits< map<T1, T2> > 
//{ static string get_typestring() { return "map< " + ::get_typestring<T1>() + ", " + ::get_typestring<T2>() + " >"; } };

} // namespace mak

#endif
